#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import ipaddress
import os
import re
import subprocess
import sys
from platform import machine

import ucpeif.loopif as loopif
import ucpeif.vxlanif as vxlif
import ucpeif.pppoeclientif as pppoeif
import ucpeif.wireguardvpnif as wireguardif
import vppapi_if.vppapi_if as vppif
from netaddr import IPNetwork, IPAddress
from pydbus import SystemBus
from pyroute2 import IPDB
from sysrepo_api import sysrepo_api3 as api
from ucpeif import ipsecif as ips
from time import sleep
from sysrepo_interfaces import _sysrepo_interfaces_wifi as wifi
from sysrepo_interfaces import radvd_configuration as radvd
import random

svc_restart_file = "/config/svcs_to_restart"
ovpn_client_file = "/config/etc/openvpn/client.conf"
rtindex_offset = 1

TVE_REBOOT_DIR = '/config/_reboot/tve/now'
sys.path.append('/scripts')

is_arm_arch = ('aarch64' in machine())

CLI_PORT = 22222
TVEAPI_PORT = 7373
SSH_PORT_FILE = '/config/ssh_port'


class IfUpError(Exception):
    def __init__(self, message, exit_code):
        super(IfUpError, self).__init__(message)
        self.exit_code = exit_code


class SysrepoInterfaces(api.SysrepoAPI):
    VXLAN_OFFSET = 500
    BR_PREFFIX = 'br-'

    def __init__(self,
                 app_name="interfaces",
                 logger=None,
                 log_file=None,
                 log_level=None,
                 console_log=False,
                 journal_log=True,
                 indexes={"": "name", "wan-interface": "name", "proposals": "id", "dh-groups": "group", "rounds": "id",
                          "wifi-wan-interface": "name", "mac-allow": "mac", "mac-deny": "mac", "probes": "id",
                          "peers": "peer-endpoint"},
                 namespace="/ucpe:config/interfaces:interface",
                 status_namespace="/ucpe:status/interfaces:interface",
                 run_as_subscriber=False,
                 priority=0):
        api.SysrepoAPI.__init__(
            self,
            app_name=app_name,
            logger=logger,
            log_file=log_file,
            log_level=log_level,
            console_log=console_log,
            journal_log=journal_log,
            namespace=namespace,
            status_namespace=status_namespace,
            indexes=indexes,
            run_as_subscriber=run_as_subscriber,
            priority=priority
        )
        self.bus = None
        self.natpassthr = None
        self.tracehost = None
        self.ipsvpn_interfaces = []
        self.vxlan_interfaces = []
        self.pppoe_interfaces = []
        self.loopback_interfaces = []
        self.wg_interfaces = []
        self.hw_type = self.get_hw_type().lower()
        self.radvd_ifs = {}
        
    def vpp_not_running(self):
        if self.__is_svc_masked('vpp-config.service'):
            return True
        if not os.path.isfile('/run/vpp.pid'):
            return True
        return False

    def get_hw_type(self):
        if not os.path.isfile('/config/.hwtype'):
            return 'intel'
        with open('/config/.hwtype', 'r') as hw_type_file:
            return hw_type_file.readline().splitlines()[0]

    def run_subscriber(self):
        try:
            self.bus = SystemBus()
            if self.bus is not None:
                self.tracehost = self.bus.get('net.telco.TracehostMonitor')
        except:
            self.logger.error('Connecting to \'net.telco.TracehostMonitor\' failed !')

        if not api.is_nanopi():
            try:
                self.bus = SystemBus()
                if self.bus is not None:
                    self.natpassthr = self.bus.get('net.telco.NatIfChange')
            except:
                self.logger.error('Connecting to \'net.telco.NatIfChange\' failed !')

        self.init_pci_data()
        api.SysrepoAPI.run_subscriber(self)

    def __get_ifname(self, name, vlan=None):
        ifname = name

        if name.find("data") > -1:
            ifname = "br-{}".format(name)

        if 'lte' in name:
            ifname = "br-{}".format(name)

        if 'wifi' in name:
            ifname = "br-{}".format(name)

        if vlan is not None and vlan != 0:
            if self.get_network_mode() == 'vpp':
                return ifname
            ifname = "{}.{}".format(ifname, vlan)
        return ifname

    get_if_sysname = __get_ifname

    def _get_ssh_port(self):
        try:
            with open(SSH_PORT_FILE, 'r') as file:
                content = file.read().strip()
                ssh_port = int(content) if content else 22222

        except FileNotFoundError:
            self.logger.error(f"Error: The file {SSH_PORT_FILE} was not found. Using default port 22222.")
            ssh_port = 22222
        except Exception as e:
            self.logger.error(f"Error occurred: {e} while getting the ssh-port.")
            ssh_port = 22222

        return ssh_port

    def _delete_nat_interface(self, name, config):
        global rtindex_offset
        # Remove IPTable config
        if 'wan-interface' in config:
            for wan in config['wan-interface']:
                wan_tag = self.get_interface_param(wan['name'], 'vlan', None)
                self._clean_nat_iptables(name, wan['name'], wan_tag=wan_tag)
        if 'wifi-wan-interface' in config:
            for wan in config['wifi-wan-interface']:
                wan_tag = self.get_interface_param(wan['name'], 'vlan', None)
                self._clean_nat_iptables(name, wan['name'], wan_tag=wan_tag)

        if 'wifi' in name:
            if '_vap' in name:
                table_id = int(name.split("_")[0].replace('wifi','')) + rtindex_offset + 5 + int(name.split("_")[1].replace("vap",""))
            else:
                table_id = int(name.replace('wifi', '')) + rtindex_offset + 5
            self._shell_return_ok("ip rule del iif br-{} table {}".format(name, table_id))
            self._shell_return_ok("ip rule del iif br-{} unreachable".format(name))
            self._shell_return_ok("ip route flush table {}".format(table_id))
            self._shell_return_ok("conntrack  -D -n")
        else:
            table_id = int(name.replace('nat', '')) + rtindex_offset
            self._shell_return_ok("ip rule del iif {} table {}".format(name, table_id))
            self._shell_return_ok("ip rule del iif {} unreachable".format(name))
            self._shell_return_ok("ip route flush table {}".format(table_id))
            self._shell_return_ok("conntrack  -D -n")
            self.execute_shell_command("/scripts/apply_networking.py -z {}".format(name))

    def __create_nat_interface(self, name, config, change_set):
        global rtindex_offset

        if 'enable' in config and config['enable']:
            if 'wan-interface' in config:
                for wan in config['wan-interface']:
                    vlan = self.get_interface_param(wan['name'], 'vlan', None)
                    self.__set_nat_iptables(name, wan['name'], wan_tag=vlan)
                    self.__set_wan_routes(name, wan['name'], wan_tag=vlan)
                    if self.natpassthr is not None:
                        self.logger.debug('Add wan interface \'{}\' to pass-thr for nat \'{}\' ' .format(wan['name'], name))
                        self.natpassthr.nat_wan_add(name, wan['name'])

            table_id = int(name.replace('nat', '')) + rtindex_offset
            self.execute_shell_command("ip rule add iif {} unreachable".format(name))
            self.execute_shell_command("ip rule add iif {} table {}".format(name, table_id))

        ipv4_prefix = self.get_interface_param(name, 'ipv4-prefix', '')
        ipv6_prefix = self.get_interface_param(name, 'ipv6-prefix', '')
        if change_set:
            for xpath, change in change_set.items():
                if change['operation'] == 'DELETED':
                    # Fill defaults
                    if xpath.find('ipv4-prefix') > -1:
                        ipv4_prefix = ''
                    if xpath.find('ipv6-prefix') > -1:
                        ipv6_prefix = ''
                    continue
                if xpath.find('ipv4-prefix') > -1:
                    ipv4_prefix = change['new_value']
                if xpath.find('ipv6-prefix') > -1:
                    ipv6_prefix = change['new_value']

        bootproto = 'manual'
        if ipv4_prefix or ipv6_prefix:
            bootproto = "static"

        options = ["-n {}".format(name), "--ovs-master ovs"]
        if bootproto:
            options.append("-b {}".format(bootproto))
        if ipv4_prefix:
            options.append("--ipv4-prefix={}".format(ipv4_prefix))
        if ipv6_prefix:
            options.append("--ipv6-prefix={}".format(ipv6_prefix))
            # Configure and enable radvd for IPv6
            self.radvd_ifs[name] = ipv6_prefix
        else:
            try:
                if name in self.radvd_ifs.keys():
                    del self.radvd_ifs[name]
            except:
                self.logger.info(f'Cannot delete radvd config for: {name}')

        self.execute_shell_command("/scripts/apply_networking.py {}".format(" ".join(options)))

    def __get_port_fwd(self):
        lck_f = open('/run/port-fwd.lck', 'a+')
        import fcntl
        import json
        try:
            fcntl.lockf(lck_f.fileno(), fcntl.LOCK_EX)
        except IOError:
            print('Upgrade procedure already running !!!')
            return {}

        if os.path.exists('/config/port-fwd.cfg'):
            jf = open('/config/port-fwd.cfg', 'r+')
            cfg = json.load(jf)
            jf.close()
        else:
            cfg = {}

        fcntl.lockf(lck_f.fileno(), fcntl.LOCK_UN)
        lck_f.close()
        return cfg

    def get_ip_mask_kernel_intfc(self, intfc):
        with IPDB() as ip:
            iface = ip.interfaces[intfc]
            ip_address = iface.ipaddr[0]['address']
            netmask = iface.ipaddr[0]['prefixlen']
            return ip_address, netmask

    def __set_nat_iptables(self, nat, wan, wan_tag=None):
        sys_wan = self.__get_ifname(wan, vlan=wan_tag)
        for iptables in {'iptables', 'ip6tables'}:
            if 'wifi' in nat:
                self._shell_return_ok("{} -A FORWARD -i {} -o br-{} -m state --state RELATED,ESTABLISHED -j ACCEPT".format(iptables, sys_wan, nat))
                self._shell_return_ok("{} -A FORWARD -i br-{} -o {} -j ACCEPT".format(iptables, nat, sys_wan))
            else:
                self._shell_return_ok("{} -A FORWARD -i {} -o {} -m state --state RELATED,ESTABLISHED -j ACCEPT".format(iptables, sys_wan, nat))
                self._shell_return_ok("{} -A FORWARD -i {} -o {} -j ACCEPT".format(iptables, nat, sys_wan))

            self._shell_return_ok("{} -t nat -A POSTROUTING -o {} -j MASQUERADE".format(iptables, sys_wan))

        if 'wifi' not in nat:
            port_fwd = self.__get_port_fwd()
            if nat in port_fwd.keys():
                port_fwd_lst = port_fwd[nat]
                for entry in port_fwd_lst:
                    dst_addr_ver = IPNetwork(entry['dst-addr']).version
                    if dst_addr_ver is 6:
                        iptables = "ip6tables"
                    else:
                        iptables = "iptables"

                    ipt_cmd = '{iptables} -t nat -A PREROUTING -i {wan_if} -p {proto} --dport {port} -j DNAT --to {dst_addr}:{dst_port}'.format(iptables=iptables, wan_if=sys_wan, proto=entry['proto'],
                                                                                                                                                port=entry['port'], dst_addr=entry['dst-addr'], dst_port=str(entry['dst-port']))
                    self._shell_return_ok(ipt_cmd)

    def _clean_nat_iptables(self, nat, wan, wan_tag=None):
        sys_wan = self.__get_ifname(wan, vlan=wan_tag)
        for iptables in {'iptables', 'ip6tables'}:
            if 'wifi' in nat:
                self._shell_return_ok("{} -D FORWARD -i {} -o br-{} -m state --state RELATED,ESTABLISHED -j ACCEPT".format(iptables, sys_wan, nat))
                self._shell_return_ok("{} -D FORWARD -i br-{} -o {} -j ACCEPT".format(iptables, nat, sys_wan))
            else:
                self._shell_return_ok("{} -D FORWARD -i {} -o {} -m state --state RELATED,ESTABLISHED -j ACCEPT".format(iptables, sys_wan, nat))
                self._shell_return_ok("{} -D FORWARD -i {} -o {} -j ACCEPT".format(iptables, nat, sys_wan))

            self._shell_return_ok("{} -t nat -D POSTROUTING -o {} -j MASQUERADE".format(iptables, sys_wan))
            self._shell_return_ok("conntrack  -D -n")

        if 'wifi' not in nat:
            port_fwd = self.__get_port_fwd()
            if nat in port_fwd.keys():
                port_fwd_lst = port_fwd[nat]
                for entry in port_fwd_lst:
                    dst_addr_ver = IPNetwork(entry['dst-addr']).version
                    if dst_addr_ver is 6:
                        iptables = "ip6tables"
                    else:
                        iptables = "iptables"

                    ipt_cmd = '{iptables} -t nat -D PREROUTING -i {wan_if} -p {proto} --dport {port} -j DNAT --to {dst_addr}:{dst_port}'.format(iptables=iptables, wan_if=sys_wan, proto=entry['proto'],
                                                                                                                                                port=entry['port'], dst_addr=entry['dst-addr'], dst_port=str(entry['dst-port']))
                    self._shell_return_ok(ipt_cmd)

    def __set_wan_routes(self, nat, wan, wan_tag=None):
        global rtindex_offset
        global is_arm_arch
        sys_wan = self.__get_ifname(wan, vlan=wan_tag)
        if 'wifi' in nat:
            if 'vap' in nat:
                table_id = int(nat.split("_")[0].replace('wifi','')) + rtindex_offset + 5 + int(nat.split("_")[1].replace("vap",""))
            else:
                table_id = int(nat.replace('wifi', '')) + rtindex_offset + 5
        else:
            table_id = int(nat.replace('nat', '')) + rtindex_offset

        ipdb = IPDB()
        from pyroute2.netlink.exceptions import NetlinkError
        if sys_wan in ipdb.interfaces:
            wan_idx = ipdb.interfaces[sys_wan]
            for rt in ipdb.routes:
                # ARM has problem with iproute2, it returns IPv4 routes as IPV6 ones
                if is_arm_arch and rt['family'] == 10:
                    continue
                if rt['oif'] == wan_idx['index']:
                    if rt['gateway']:
                        if rt['gateway'].startswith("fe80:") : #link-local address
                            route = {'dst': rt['dst'], 'gateway': rt['gateway'], 'table': table_id, 'priority': rt['priority'], 'oif': wan_idx['index']}
                        else:
                            route = {'dst': rt['dst'], 'gateway': rt['gateway'], 'table': table_id, 'priority': rt['priority']}
                    else:
                        route = {'dst': rt['dst'], 'table': table_id, 'oif': wan_idx['index']}

                    try:
                        ipdb.routes.add(route).commit()
                    except NetlinkError as e:
                        self.logger.error('Set route in table {tbl} failed. {msg}'.format(tbl=table_id, msg=str(e)))
                        if e.code != 17:
                            self.logger.error('Set route in table {tbl} failed. {msg}'.format(tbl=table_id, msg=str(e)))
                    if os.path.isfile("/config/nat-lb"):
                        os.system("ip route del  default  metric 1  table %d" % table_id)
                        cmd='ip route add default proto static metric 1 table %d $(for metric in `seq 2 249` ; do  ip route list default table %d metric $metric ; done | sed \'s/default/nexthop/\'  | sed \'s/proto static/weight 1/\' | tr "\n" " ")' %  (table_id, table_id)
                        os.system(cmd)

        ipdb.release()

    def __clean_wan_routes(self, nat, wan, wan_tag=None):
        global rtindex_offset
        sys_wan = self.__get_ifname(wan, vlan=wan_tag)
        if 'wifi' in nat:
            if 'vap' in nat:
                table_id = int(nat.split("_")[0].replace('wifi','')) + rtindex_offset + 5 + int(nat.split("_")[1].replace("vap",""))
            else:
                table_id = int(nat.replace('wifi', '')) + rtindex_offset + 5
        else:
            table_id = int(nat.replace('nat', '')) + rtindex_offset

        ipdb = IPDB()
        if sys_wan in ipdb.interfaces:
            wan_idx = ipdb.interfaces[sys_wan]
            if table_id in ipdb.routes.tables.keys():
                for rt in ipdb.routes.tables[table_id]:
                    dst = "none"
                    if 'dst' in rt:
                        dst = rt['dst']
                    if rt['oif'] == wan_idx['index']:
                        try:
                            with ipdb.routes.tables[str(table_id)][rt['dst']] as r:
                                r.remove()
                        except:
                            self.logger.error("Can't delete route %s from table %s" % (dst,str(table_id)))
                    if os.path.isfile("/config/nat-lb"):
                        os.system("ip route del  default  metric 1  table %d" % table_id)

    def __config_mgmt_interface(self, name, config, change_set):

        bootproto = self.get_interface_param(name, 'bootproto', 'dhcp')
        ipv4_prefix = self.get_interface_param(name, 'ipv4-prefix', '')
        ipv6_prefix = self.get_interface_param(name, 'ipv6-prefix', '')

        if change_set:
            for xpath, change in change_set.items():
                if change['operation'] == 'DELETED':
                    # Fill defaults
                    if xpath.find('bootproto') > -1:
                        bootproto = ''
                    if xpath.find('ipv4-prefix') > -1:
                        ipv4_prefix = ''
                    if xpath.find('ipv6-prefix') > -1:
                        ipv6_prefix = ''
                    if xpath.find('dns') > -1:
                        dns = ''
                    if xpath.find('dns6') > -1:
                        dns6 = ''
                    if xpath.find('metric') > -1:
                        metric = 100
                    if xpath.find('vlan') > -1:
                        vlan = 0
                    if xpath.find('gateway6') > -1:
                        gateway6 = ''
                    elif xpath.find('gateway') > -1:
                        gateway = ''
                    if xpath.find('management') > -1:
                        management = 'false'
                    continue
                if xpath.find('ipv4-prefix') > -1:
                    ipv4_prefix = change['new_value']
                if xpath.find('ipv6-prefix') > -1:
                    ipv6_prefix = change['new_value']

        if ipv4_prefix or ipv6_prefix:
            bootproto = "static"
        else:
            bootproto = "manual"

        self._shell_return_ok("ifdown --ignore-error --force {}".format(name))
        self._shell_return_ok("rm -rf /etc/network/interfaces.d/ifcfg-{}".format(name))
        onbt = "yes"
        options = ["-n {}".format(name), "-o {}".format(onbt)]
        options.append("-b {}".format(bootproto))
        if ipv4_prefix:
            options.append("--ipv4-prefix={}".format(ipv4_prefix))
        if ipv6_prefix:
            options.append("--ipv6-prefix={}".format(ipv6_prefix))

        options.append("--ovs-master ovs --ovs-ports {}".format(name))
        self.execute_shell_command("/scripts/apply_networking.py {}".format(" ".join(options)))

        # Configure and enable RA for IPv6
        if ipv6_prefix:
            self.radvd_ifs[name] = ipv6_prefix
        else:
            try:
                if name in self.radvd_ifs.keys():
                    del self.radvd_ifs[name]
            except:
                self.logger.info(f'Cannot delete radvd config for: {name}')

        if onbt == "yes":
            try:
                self.logger.info("ifup interface {}".format(name))
                subprocess.check_call(['ifup', '--ignore-error', name])
                if str(self.execute_shell_command("systemctl is-active isc-dhcp-relay")).strip("\n") == "active":
                    from sysrepo_dhcprelay import sysrepo_dhcprelay as dhcp_rel_api
                    dhcprelay = dhcp_rel_api.SysrepoDhcpRelay(app_name="sysrepo_interfaces")
                    if dhcprelay.get_relay_int() == "mgmt0":
                        self.execute_shell_command("systemctl --no-block restart isc-dhcp-relay.service")
            except subprocess.CalledProcessError as ex:
                self.logger.error('ifup command failed!')
                self._shell_return_ok("ifdown --ignore-error --force {}".format(name))
                raise IfUpError('ifup error: {}'.format(ex), 1)
            except ImportError as e:
                self.logger.critical('Fail module import {}'.format(e))
                pass

    def __config_nat_interface(self, name, config, change_set):
        old_enable_nat = False
        if config and 'enable' in config:
            old_enable_nat = config['enable']

        if not config:
            config = {}
        if 'wan-interface' not in config:
            config['wan-interface'] = []
        if change_set:
            for xpath, change in change_set.items():
                if xpath.find('enable') > -1 and change['operation'] != 'DELETED':
                    config['enable'] = change['new_value']
                    if 'true' in change['new_value']:
                        config['enable'] = True
                    elif 'false' in change['new_value']:
                        config['enable'] = False
                if xpath.find('wan-interface') > -1:
                    if change['operation'] == 'DELETED':
                        if not change['old_value']:
                            continue
                        index = 0
                        for wan in config['wan-interface']:
                            if wan['name'] == change['old_value']:
                                wan_tag = self.get_interface_param(wan['name'], 'vlan', None)
                                self._clean_nat_iptables(name, wan['name'], wan_tag=wan_tag)
                                config['wan-interface'].pop(index)
                                if self.natpassthr is not None:
                                    self.logger.debug('Delete wan interface \'{}\' from pass-thr for nat \'{}\' ' .format(wan['name'], name))
                                    self.natpassthr.nat_wan_del(name, wan['name'])
                                break
                            index = index + 1
                    else:
                        if not change['new_value']:
                            continue
                        config['wan-interface'].append({'name': change['new_value']})
                if xpath.find('port-forwarding-all') > -1 and change['operation'] != 'DELETED':
                    config['port-forwarding-all'] = change['new_value']
                    if 'true' == change['new_value']:
                        config['port-forwarding-all'] = True
                    elif 'false' == change['new_value']:
                        config['port-forwarding-all'] = False
        if old_enable_nat:
            self._shell_return_ok("ifdown --ignore-error {}".format(name))
            self._delete_nat_interface(name, config)

        self.__create_nat_interface(name, config, change_set)
        if 'enable' in config and config['enable']:
            self._shell_return_ok("ifup --ignore-error {}".format(name))

        if config and 'enable' in config and self.natpassthr is not None:
            self.logger.debug('Update nat interface state \'{}\' to pass-thr for nat \'{}\' ' .format(config['enable'], name))
            self.natpassthr.nat_enable(name, config['enable'])

            if 'port-forwarding-all' in config:
                self.logger.debug('Update pass-thr state \'{}\' to pass-thr for nat \'{}\' ' .format(config['port-forwarding-all'], name))
                self.natpassthr.nat_passthr(name, config['port-forwarding-all'])

    def __set_ippass_enable_udp_dst_port(self, port, enable):
        if_cfg = self.get_interfaces()
        ippass_ifs = list(filter(lambda x: True if (x.get('type', '') == 'ethernet' or x.get('type','') == 'lte' or x.get('type','') == 'wifi'  or x.get('type','') == 'pppoe') and x.get('ip-passthrough', False) else False, list(if_cfg.values())))

        if not ippass_ifs or self.vpp_not_running():
            return

        vpp_if = vppif.VppApiInterface('sr-interfaces')
        for interface in ippass_ifs:
            ifname = interface.get('name', 'noname')
            ifvlan = interface.get('vlan', 0)
            if ifvlan != 0:
                ifname += '.' + str(ifvlan)

            (if_name, vpp_idx) = vpp_if.search_vppif_by_hostifname(ifname)
            if if_name == None or vpp_idx == None:
                continue

            ippass_host_idx = vpp_if.ippass_get_host_if_idx(vpp_idx)
            ret = vpp_if.add_del_ippass_permit_host_entry(ippass_host_idx, 'udp', dst_port=int(port),  add=1 if enable else 0)
            self.logger.debug('add_del_ippass_permit_host_entry(ippass_host_idx {}, \'udp\', dst_port={},  add={}, ret {})'.format(ippass_host_idx, port, 1 if enable else 0, ret))
        del(vpp_if)

    def __set_ippass_enable_esp_proto(self, enable):
        if_cfg = self.get_interfaces()
        ippass_ifs = list(filter(lambda x: True if (x.get('type', '') == 'ethernet' or x.get('type','') == 'lte' or x.get('type','') == 'wifi' or x.get('type','') == 'pppoe' ) and x.get('ip-passthrough', False) else False, list(if_cfg.values())))

        if not ippass_ifs or self.vpp_not_running():
            return

        vpp_if = vppif.VppApiInterface('sr-interfaces')
        for interface in ippass_ifs:
            (if_name, vpp_idx) = vpp_if.search_vppif_by_hostifname(interface.get('name', 'noname'))
            if if_name == None or vpp_idx == None:
                continue

            ippass_host_idx = vpp_if.ippass_get_host_if_idx(vpp_idx)
            ret = vpp_if.add_del_ippass_permit_host_entry(ippass_host_idx, 'esp', add=1 if enable else 0)
            self.logger.debug('add_del_ippass_permit_host_entry(ippass_host_idx {}, \'esp\', add={}, ret {})'.format(ippass_host_idx, 1 if enable else 0, ret))
        del(vpp_if)

    def __add_del_ip_pass_tap_intf_if_wg_enabled(self, ifname, add_del):
        if_cfg = self.get_interfaces()
        ippass_ifs = list(filter(lambda x: True if (x.get('type', '') == 'ethernet' or x.get('type', '') == 'lte' or x.get(
            'type', '') == 'wifi' or x.get('type', '') == 'pppoe') and x.get('ip-passthrough', False) else False, list(if_cfg.values())))

        local_endpoint = self.get_interface_param(ifname, 'local-endpoint', '')

        vpp_if = vppif.VppApiInterface('sr-interfaces')
        for interface in ippass_ifs:
            intfname = interface.get('name', 'noname')
            ippass_prefix = self.get_ip_mask_kernel_intfc(f'br-{intfname}')

            if local_endpoint == ippass_prefix[0]:
                local_port = self.get_interface_param(ifname, 'local-port', '')
                peer_endpoint = self.get_wireguard_peer_params(ifname, 'peer-endpoint')

                if local_endpoint and local_port and peer_endpoint:

                    (if_name, vpp_idx) = vpp_if.search_vppif_by_hostifname(intfname)
                    ippass_host_idx = vpp_if.ippass_get_host_if_idx(vpp_idx)
                    bridge_data = vpp_if.get_interface_bd(ippass_host_idx)

                    sw_if_index_values = []
                    for sw_if_detail in bridge_data['sw_if_details']:
                        sw_if_index_values.append(sw_if_detail['sw_if_index'])

                    for idx in sw_if_index_values:
                        intf_name = vpp_if.get_ifname_by_ifidx(idx)
                        if intf_name.startswith('tap'):
                            ret = vpp_if.add_del_ippass_permit_host_entry(idx, 'udp', 0, int(local_port), local_endpoint,
                                                                          peer_endpoint, add_del)

                            vpp_if.add_del_ippass_permit_host_entry(host_idx=idx, protocol='tcp', dst_port=7374)
                            vpp_if.add_del_ippass_permit_host_entry(host_idx=idx, protocol='tcp', dst_port=7373)
                            break

    def __config_openvpn_ippass(self, old_port, new_port, old_enable, new_enable, startup):
        self.logger.debug('__config_openvpn_ippass({},{},{},{}, {})'.format(old_port, new_port, old_enable, new_enable, startup))
        if startup and new_enable:
            self.logger.debug('OVPN startup config - {} {}'.format(new_enable, new_port))
            self.__set_ippass_enable_udp_dst_port(new_port, new_enable)

        if not new_enable and old_enable:
            # OVPN have to be disabled
            # remove entries for old port
            self.logger.debug('OVPN is disabled - remove entries for port {}'.format(old_port))
            self.__set_ippass_enable_udp_dst_port(old_port, new_enable)

        if not old_enable and new_enable:
            # OVPN is enabled
            # add entries for new_port
            self.logger.debug('OVPN is enabled - add entries for port {}'.format(new_port))
            self.__set_ippass_enable_udp_dst_port(new_port, new_enable)

        if old_enable and new_enable:
            if new_port != old_port:
                # port is changed
                self.logger.debug('OVPN is enabled - remove entries for port {}'.format(old_port))
                self.__set_ippass_enable_udp_dst_port(old_port, False)
                self.logger.debug('OVPN is enabled - add entries for port {}'.format(new_port))
                self.__set_ippass_enable_udp_dst_port(new_port, True)

    def __set_ippass_enable_vxlan(self, ifname, lvtep, rvteps, tun_type, enable):

        if self.vpp_not_running():
            return

        vpp_if = vppif.VppApiInterface('sr-interfaces')
        mode = self.get_item("/ucpe:config/interfaces:interface[name='{}']/mode".format(ifname))
        if mode == 'virtio':
            if_prefix = 'host-'
        else:
            if_prefix = ''

        vpp_idx = vpp_if.get_ifidx_by_name('{}{}'.format(if_prefix, ifname))
        if not vpp_idx:
            del(vpp_if)
            return

        ippass_host_idx = vpp_if.ippass_get_host_if_idx(vpp_idx)
        for rvtep in rvteps:
            if type(rvtep) != type(lvtep):
                continue
            ret = vpp_if.add_del_ippass_permit_host_entry(host_idx=ippass_host_idx,
                                                          protocol='udp',
                                                          dir_out=0,
                                                          dst_port=4790 if tun_type == 'vxlan-gpe' else 4789,
                                                          ip=lvtep,
                                                          src_ip=rvtep,
                                                          add=1 if enable else 0)
            self.logger.debug('add_del_ippass_permit_host_entry ret {})'.format(ret))
        del(vpp_if)

    def __get_vxlan_lvtep_ippass(self, lvtep):
        if_cfg = self.get_interfaces()
        ippass_ifs = list(filter(lambda x: True if (x.get('type', '') == 'ethernet' or x.get('type','') == 'lte' or x.get('type','') == 'wifi' or x.get('type','') == 'pppoe' ) and x.get('ip-passthrough', False) else False, list(if_cfg.values())))
        for interface in ippass_ifs:
            if lvtep.version == 6:
                address = interface.get('ipv6-prefix', None)
            else:
                address = interface.get('ipv4-prefix', None)
            if address is not None and ipaddress.ip_interface(address).ip == lvtep:
                return (interface.get('name', None), True)

        return (None, False)

    def __config_ippass_vxlan(self, old_local, old_remote, old_type, new_local=None, new_remote=[], new_type=None, startup=False):
        """ Have to find the outgoing interface from the local VTEP address confguration
                and to check if ip-passthrough is enabled.
            Then according to the Local + Remotes to add permit entry in ip-pass config to match:
                IPv4 / IPv6 UPD packets
                Dst ip - local VTEP
                Src ip - Remote VTEP (for every remote address in list)
                Dst port 4789 for vxlan ot 4790 for vxlan-gpe
        """
        if not old_remote and not new_remote and old_local == None and new_local == None:
            return
        self.logger.debug('__config_ippass_vxlan({},{},{},{},{},{},{})'.format(old_local, old_remote, old_type, new_local, new_remote, new_type, startup))

        old_ippass_enabled = False
        new_ippass_enabled = False
        old_ifname = None
        new_ifname = None

        if old_local:
            (old_ifname, old_ippass_enabled) = self.__get_vxlan_lvtep_ippass(old_local)
        if new_local:
            (new_ifname, new_ippass_enabled) = self.__get_vxlan_lvtep_ippass(new_local)

        if startup == True:
            if old_ippass_enabled and old_local and old_remote and old_type:
                self.logger.debug('VxLAN startup config - {} {} {} {}'.format(old_ifname, old_local, old_remote, old_type))
                self.__set_ippass_enable_vxlan(old_ifname, old_local, old_remote, old_type, True)
        else:
            if old_ippass_enabled and old_local and old_remote and old_type:
                self.logger.debug('VxLAN running remove config - {} {} {} {}'.format(old_ifname, old_local, old_remote, old_type))
                self.__set_ippass_enable_vxlan(old_ifname, old_local, old_remote, old_type, False)

            if new_ippass_enabled or old_ippass_enabled:
                ifname = None
                ifname = new_ifname if new_ippass_enabled else old_ifname

                if new_local and new_remote and new_type:
                    self.logger.debug('VxLAN running apply config - {} {} {} {}'.format(ifname, new_local, new_remote, new_type))
                    self.__set_ippass_enable_vxlan(ifname, new_local, new_remote, new_type, True)

    def _create_wifi_nat_interface(self, name, config, change_set):
        global rtindex_offset

        if 'nat' in config and config['nat']:
            if 'wifi-wan-interface' in config:
                for wan in config['wifi-wan-interface']:
                    vlan = self.get_interface_param(wan['name'], 'vlan', None)
                    self.__set_nat_iptables(name, wan['name'], wan_tag=vlan)
                    self.__set_wan_routes(name, wan['name'], wan_tag=vlan)
                    if self.natpassthr is not None:
                        self.logger.debug('Add wan interface \'{}\' to pass-thr for nat \'br-{}\' ' .format(wan['name'], name))
                        self.natpassthr.nat_wan_add("br-{}".format(name), wan['name'])
            if 'vap' in name:
                table_id = int(name.split("_")[0].replace('wifi','')) + rtindex_offset + 5 + int(name.split("_")[1].replace("vap",""))
            else:
                table_id = int(name.replace('wifi', '')) + rtindex_offset + 5
            self.execute_shell_command("ip rule add iif br-{} unreachable".format(name))
            self.execute_shell_command("ip rule add iif br-{} table {}".format(name, table_id))

    def __config_openvpn_interface(self, name, config, change_set):
        global ovpn_client_file

        if not config:
            self.logger.debug('{} no config'.format(name))
            return

        self.execute_shell_command("/scripts/apply_networking.py -z {}".format(name))
        enable_vpn = False
        if 'enable' in config:
            enable_vpn = config['enable']

        old_enable_vpn = enable_vpn

        concentrator_endpoint = None
        concentrator_port = None
        if 'ovpn-cfg' in config:
            if 'concentrator-endpoint' in config['ovpn-cfg']:
                concentrator_endpoint = config['ovpn-cfg']['concentrator-endpoint']
            if 'concentrator-port' in config['ovpn-cfg']:
                concentrator_port = config['ovpn-cfg']['concentrator-port']

        old_port = concentrator_port
        old_enable = old_enable_vpn

        if change_set:
            for xpath, change in change_set.items():
                if change['operation'] == 'DELETED':
                    continue
                if xpath.find('enable') > -1:
                    enable_vpn = change['new_value']
                if xpath.find('concentrator-endpoint') > -1:
                    concentrator_endpoint = change['new_value']
                if xpath.find('concentrator-port') > -1:
                    concentrator_port = change['new_value']
                    
        #get ports from current config of openvpn
        try:
            with open (ovpn_client_file, "rt") as openvpnf:
                vpncfg=openvpnf.read()
                matches=re.search("\nremote (.*) (.*)\n",vpncfg)
                if matches:
                    curr_conf_conc=(matches.group(1)).strip()
                    curr_conf_port=(matches.group(2)).strip()
                    if old_port:
                        old_port = curr_conf_port
        except:
            curr_conf_conc=None
            curr_conf_port=None
            
        #concentrator port can be a list
        if concentrator_port:
            concentrator_portslist = sum([ list(range(int(x.split("-")[0]),int(x.split("-")[1])+1)) for x in concentrator_port.split(",") if "-" in x],
                                         [ int(x) for x in concentrator_port.split(",") if "-" not in x  ])
            concentrator_port = random.choice(concentrator_portslist)
            port_in_list =  [ele for ele in concentrator_portslist if(str(ele) == str(curr_conf_port))]
            if bool(port_in_list):
                concentrator_port = curr_conf_port

        new_port = concentrator_port
        new_enable = enable_vpn
        self.__config_openvpn_ippass(old_port, new_port, old_enable, new_enable, False if change_set else True)

        if concentrator_port and concentrator_endpoint:
            if concentrator_endpoint.startswith("tcp6:"):
                concentrator_endpoint=concentrator_endpoint.replace("tcp6:","")
                self.execute_shell_command("sed -i 's/^proto .*/proto tcp6-client/' %s" % ovpn_client_file)
                self.execute_shell_command("sed -i 's/^fragment .*/#fragment 1300/' %s" % ovpn_client_file)
            elif concentrator_endpoint.startswith("tcp:"):
                concentrator_endpoint=concentrator_endpoint.replace("tcp:","")
                self.execute_shell_command("sed -i 's/^proto .*/proto tcp-client/' %s" % ovpn_client_file)
                self.execute_shell_command("sed -i 's/^fragment .*/#fragment 1300/' %s" % ovpn_client_file)
            elif concentrator_endpoint.startswith("udp6:"):
                concentrator_endpoint=concentrator_endpoint.replace("udp6:","")
                self.execute_shell_command("sed -i 's/^proto .*/proto udp6/' %s" % ovpn_client_file)
                self.execute_shell_command("sed -i 's/^#fragment .*/fragment 1300/' %s" % ovpn_client_file)
            else:
                concentrator_endpoint=concentrator_endpoint.replace("udp:","")
                concentrator_endpoint=concentrator_endpoint.replace("udp4:","")
                self.execute_shell_command("sed -i 's/^proto .*/proto udp/' %s" % ovpn_client_file)
                self.execute_shell_command("sed -i 's/^#fragment .*/fragment 1300/' %s" % ovpn_client_file)

            self.execute_shell_command("sed -i 's/^remote.*/remote\ %s\ %s/' %s" % (
                concentrator_endpoint, concentrator_port, ovpn_client_file))
            self.execute_shell_command("sed -i 's/^lport.*/lport\ %s/' %s" % (
                concentrator_port, ovpn_client_file))

        if old_enable_vpn and enable_vpn:
            self.__restart_openvpn()
        else:
            if enable_vpn:
                self.__start_openvpn()
            else:
                self.__stop_openvpn()

    def __ipsecvpn_disconnect(self, ipsvpn):
        ret = False
        if ipsvpn is None:
            return
        try:
            ret = ipsvpn.disconnect()
            if ret is False:
                self.logger.error('Disconnect IPsecVPN interface \'{}\' failed'.format(ipsvpn.name))

            ipsvpn.unload_connection()
            ipsvpn.unload_security()
        except:
            self.logger.error('Failed to disconnect IPsecVPN interface')

    def __ipsecvpn_connect(self, ipsvpn):
        ret = False
        if ipsvpn is None:
            return
        try:
            ipsvpn.load_security()
            ipsvpn.load_connection(debug=True)
            ret = ipsvpn.connect(timeout=10000)
            if ret is False:
                self.logger.error('Connect IPsecVPN interface \'{}\' failed'.format(ipsvpn.name))
        except:
            self.logger.error('Failed to connect IPsecVPN interface')

    def __set_ippass_ipsec(self, old_enable, new_enable):
        if old_enable == new_enable:
            return

        self.logger.info('IPSEC is enabled {} - set IPPass'.format(new_enable))

        self.__set_ippass_enable_udp_dst_port(500, new_enable)
        self.__set_ippass_enable_udp_dst_port(4500, new_enable)
        self.__set_ippass_enable_esp_proto(new_enable)

    def __config_ipsecvpn_startup(self, name, config, ipsvpn_interfaces):
        if not config:
            return

        for ipsvpn in ipsvpn_interfaces:
            if ipsvpn.name == name:
                for key, value in config.items():
                    if key == "ike-version":
                        if value == '1':
                            ipsvpn.ike = ips.IKE.V1
                        elif value == '2':
                            ipsvpn.ike = ips.IKE.V2
                    if key == "ike-mode":
                        if value == "main":
                            ipsvpn.ike1_mode = ips.IKE1_MODE.MAIN
                        elif value == "aggressive":
                            ipsvpn.ike1_mode = ips.IKE1_MODE.AGGRESIVE
                    elif key == "ike-phase-1":
                        if 'dh-groups' in config[key]:
                            for group in config[key]['dh-groups']:
                                ipsvpn.dh_group_add(int(group['group']), ips.IKE_PHASE.PHASE1)
                        if 'proposals' in config[key]:
                            for prop in config[key]['proposals']:
                                if 'encryption-algorithm' in prop and 'authentication-algorithm' in prop:
                                    ipsvpn.proposal_add(str(prop['encryption-algorithm']), str(prop['authentication-algorithm']), ips.IKE_PHASE.PHASE1)
                        if 'local' in config[key] and 'rounds' in config[key]['local']:
                            for auth_method in config[key]['local']['rounds']:
                                if auth_method['authentication-method'] == 'pre-shared-keys':
                                    ipsvpn.auth_add(ips.AUTH_SIDE.LOCAL, auth_method['id'], ips.IPSecAuth(method=ips.AUTH_METHOD.PSK, psk=auth_method['secret']))
                                elif auth_method['authentication-method'] == 'pubkeys':
                                    passphrase = None if 'key-pass' not in auth_method.keys() else str(auth_method['key-pass'])
                                    ipsvpn.auth_add(ips.AUTH_SIDE.LOCAL, auth_method['id'], ips.IPSecAuth(method=ips.AUTH_METHOD.PUBKEY, ca_crt=auth_method['ca-cert'], crt=auth_method['cert'], key=auth_method['key'], passphrase=passphrase))
                                elif auth_method['authentication-method'] == 'xauth':
                                    ipsvpn.auth_add(ips.AUTH_SIDE.LOCAL, auth_method['id'], ips.IPSecAuth(method=ips.AUTH_METHOD.XAUTH, user=auth_method['username'], psw=auth_method['password']))
                                elif auth_method['authentication-method'] == 'eap':
                                    passphrase = None if 'key-pass' not in auth_method.keys() else str(auth_method['key-pass'])
                                    ipsvpn.auth_add(ips.AUTH_SIDE.LOCAL, auth_method['id'], ips.IPSecAuth(method=ips.AUTH_METHOD.EAP, user=auth_method['username'], psw=auth_method['password'], ca_crt=auth_method['ca-cert'], crt=auth_method['cert'], key=auth_method['key'], passphrase=passphrase))
                        if 'remote' in config[key] and 'rounds' in config[key]['remote']:
                            for auth_method in config[key]['remote']['rounds']:
                                if auth_method['authentication-method'] == 'pre-shared-keys':
                                    ipsvpn.auth_add(ips.AUTH_SIDE.REMOTE, auth_method['id'], ips.IPSecAuth(method=ips.AUTH_METHOD.PSK, psk=auth_method['secret']))
                                elif auth_method['authentication-method'] == 'pubkeys':
                                    ipsvpn.auth_add(ips.AUTH_SIDE.REMOTE, auth_method['id'], ips.IPSecAuth(method=ips.AUTH_METHOD.PUBKEY, ca_crt=auth_method['ca-cert'], crt=auth_method['cert']))
                    elif key == "ike-phase-2":
                        if 'dh-groups' in config[key]:
                            for group in config[key]['dh-groups']:
                                ipsvpn.dh_group_add(int(group['group']), ips.IKE_PHASE.PHASE2)
                        if 'proposals' in config[key]:
                            for prop in config[key]['proposals']:
                                if 'encryption-algorithm' in prop and 'authentication-algorithm' in prop:
                                    ipsvpn.proposal_add(str(prop['encryption-algorithm']), str(prop['authentication-algorithm']), ips.IKE_PHASE.PHASE2)
                        if 'local-ts' in config[key] and config[key]['local-ts']:
                            ipsvpn.local_ts = IPNetwork(config[key]['local-ts'])
                        if 'remote-ts' in config[key] and config[key]['remote-ts']:
                            ipsvpn.remote_ts = IPNetwork(config[key]['remote-ts'])
                    elif key == 'remote-gw':
                        ipsvpn.gateway = str(value)
                    elif key == 'local-id':
                        ipsvpn.local_id = str(value)
                    elif key == 'remote-id':
                        ipsvpn.remote_id = str(value)
                    elif key == 'config-mode':
                        if value == 'pull':
                            ipsvpn.mode_config = ips.MODE_CONFIG.PULL
                        elif value == 'push':
                            ipsvpn.mode_config = ips.MODE_CONFIG.PUSH
                    elif key == 'protocol-mode':
                        if value == 'tunnel':
                            ipsvpn.child_mode = ips.CHILD_MODE.TUNNEL
                        elif value == 'transport':
                            ipsvpn.child_mode = ips.CHILD_MODE.TRANSPORT
                    elif key == 'install-virtual-ip':
                        if value is False or value == 'false':
                            ipsvpn.vips = None
                        else:
                            ipsvpn.vips = IPAddress(config.get('virtual-ip', '0.0.0.0'))

                if 'enable' in config and (config['enable'] is True or config['enable'] == 'true'):
                    self.__set_ippass_ipsec(False, config.get('enable', False))
                    self.__ipsecvpn_connect(ipsvpn)

    def __config_ipsecvpn_apply(self, name, change_set, ipsvpn_interfaces):
        if not change_set:
            return

        rounds = {'local': [{}, {}], 'remote': [{}]}
        props = {'ike-phase-1': [{}, {}, {}, {}, {}], 'ike-phase-2': [{}, {}, {}, {}, {}]}
        old_props = {'ike-phase-1': [{}, {}, {}, {}, {}], 'ike-phase-2': [{}, {}, {}, {}, {}]}
        l_rounds = self.get_ike1_local_rounds(name)
        if l_rounds:
            for l in l_rounds:
                rounds['local'][l['id'] - 1] = l

        r_rounds = self.get_ike1_remote_rounds(name)
        if r_rounds:
            for r in r_rounds:
                rounds['remote'][r['id'] - 1] = r

        for ipsvpn in ipsvpn_interfaces:
            if ipsvpn.name == name:
                if_status = self.get_interface_param(name, 'enable')
                inst_vip = self.get_install_vip(name)
                vip = self.get_vip(name)
                prev_status = if_status
                for xpath, change in change_set.items():
                    # Get proposal and round IDs
                    prop_id = change['xpath'][change['xpath'].find('proposals[id=\'')+len('proposals[id=\''):change['xpath'].rfind('\'')]
                    round_id = change['xpath'][change['xpath'].find('rounds[id=\'')+len('rounds[id=\''):change['xpath'].rfind('\'')]

                    if change['operation'] == 'DELETED':
                        if xpath.find('remote-gw') > -1:
                            ipsvpn.gateway = None
                        if xpath.find('local-id') > -1:
                            del ipsvpn.local_id
                        if xpath.find('remote-id') > -1:
                            del ipsvpn.remote_id
                        if xpath.find('ike-version') > -1:
                            ipsvpn.ike = ips.IKE.V2
                        if xpath.find('ike-mode') > -1:
                            ipsvpn.ike1_mode = ips.IKE1_MODE.MAIN
                        if xpath.find('rounds') > -1:
                            if xpath.find('local/rounds') > -1 and xpath.endswith('\']'):
                                rounds['local'][int(round_id) - 1] = {}
                                ipsvpn.auth_del(ips.AUTH_SIDE.LOCAL, int(round_id))
                            elif xpath.find('remote/rounds') > -1 and xpath.endswith('\']'):
                                rounds['remote'][int(round_id) - 1] = {}
                                ipsvpn.auth_del(ips.AUTH_SIDE.REMOTE, int(round_id))
                            elif xpath.find('local') > -1 and xpath.find('key-pass') > -1:
                                rounds['local'][int(round_id) - 1]['key-pass'] = None
                                rounds['local'][int(round_id) - 1]['modified'] = True
                        if xpath.find('ike-phase-1/dh-group') > -1:
                            dh_group = change['old_value']
                            if dh_group:
                                ipsvpn.dh_group_del(int(dh_group), ips.IKE_PHASE.PHASE1)
                        if xpath.find('ike-phase-2/dh-group') > -1:
                            dh_group = change['old_value']
                            if dh_group:
                                ipsvpn.dh_group_del(int(dh_group), ips.IKE_PHASE.PHASE2)
                        if xpath.find('ike-phase-1/proposals') > -1:
                            if xpath.find('encryption-algorithm') > -1:
                                old_props['ike-phase-1'][int(prop_id) - 1]['id'] = prop_id
                                old_props['ike-phase-1'][int(prop_id) - 1]['enc'] = change['old_value']
                            if xpath.find('authentication-algorithm') > -1:
                                old_props['ike-phase-1'][int(prop_id) - 1]['id'] = prop_id
                                old_props['ike-phase-1'][int(prop_id) - 1]['auth'] = change['old_value']
                        if xpath.find('ike-phase-2/proposals') > -1:
                            if xpath.find('encryption-algorithm') > -1:
                                old_props['ike-phase-2'][int(prop_id) - 1]['id'] = prop_id
                                old_props['ike-phase-2'][int(prop_id) - 1]['enc'] = change['old_value']
                            if xpath.find('authentication-algorithm') > -1:
                                old_props['ike-phase-2'][int(prop_id) - 1]['id'] = prop_id
                                old_props['ike-phase-2'][int(prop_id) - 1]['auth'] = change['old_value']
                        if xpath.find('local-ts') > -1:
                            del ipsvpn.local_ts
                        if xpath.find('remote-ts') > -1:
                            del ipsvpn.remote_ts
                        continue

                    if xpath.find('remote-gw') > -1:
                        gateway = change['new_value']
                        ipsvpn.gateway = gateway
                    if xpath.find('local-id') > -1:
                        local_id = change['new_value']
                        if local_id == 'None' or local_id == 'none':
                            del ipsvpn.local_id
                        else:
                            ipsvpn.local_id = local_id
                    if xpath.find('remote-id') > -1:
                        remote_id = change['new_value']
                        if remote_id == 'None' or remote_id == 'none':
                            del ipsvpn.remote_id
                        else:
                            ipsvpn.remote_id = remote_id
                    if xpath.find('config-mode') > -1:
                        if change['new_value'] == 'pull':
                            ipsvpn.mode_config = ips.MODE_CONFIG.PULL
                        elif change['new_value'] == 'push':
                            ipsvpn.mode_config = ips.MODE_CONFIG.PUSH
                    if xpath.find('protocol-mode') > -1:
                        if change['new_value'] == 'tunnel':
                            ipsvpn.child_mode = ips.CHILD_MODE.TUNNEL
                        elif change['new_value'] == 'transport':
                            ipsvpn.child_mode = ips.CHILD_MODE.TRANSPORT
                    if xpath.find('ike-version') > -1:
                        ike_version = change['new_value']
                        if ike_version == '1':
                            ipsvpn.ike = ips.IKE.V1
                        elif ike_version == '2':
                            ipsvpn.ike = ips.IKE.V2
                    if xpath.find('ike-mode') > -1:
                        ike_mode = change['new_value']
                        if ike_mode == 'main':
                            ipsvpn.ike1_mode = ips.IKE1_MODE.MAIN
                        elif ike_mode == 'aggressive':
                            ipsvpn.ike1_mode = ips.IKE1_MODE.AGGRESIVE
                    if xpath.find('ike-phase-1/proposals') > -1:
                        if xpath.find('encryption-algorithm') > -1:
                            props['ike-phase-1'][int(prop_id) - 1]['id'] = prop_id
                            props['ike-phase-1'][int(prop_id) - 1]['enc'] = change['new_value']
                        if xpath.find('authentication-algorithm') > -1:
                            props['ike-phase-1'][int(prop_id) - 1]['id'] = prop_id
                            props['ike-phase-1'][int(prop_id) - 1]['auth'] = change['new_value']
                    if xpath.find('ike-phase-2/proposals') > -1:
                        if xpath.find('encryption-algorithm') > -1:
                            props['ike-phase-2'][int(prop_id) - 1]['id'] = prop_id
                            props['ike-phase-2'][int(prop_id) - 1]['enc'] = change['new_value']

                        if xpath.find('authentication-algorithm') > -1:
                            props['ike-phase-2'][int(prop_id) - 1]['id'] = prop_id
                            props['ike-phase-2'][int(prop_id) - 1]['auth'] = change['new_value']

                    if xpath.find('ike-phase-1/dh-group') > -1:
                        dh_group = change['new_value']
                        if dh_group:
                            ipsvpn.dh_group_add(int(dh_group), ips.IKE_PHASE.PHASE1)

                    if xpath.find('ike-phase-2/dh-group') > -1:
                        dh_group = change['new_value']
                        if dh_group:
                            ipsvpn.dh_group_add(int(dh_group), ips.IKE_PHASE.PHASE2)

                    if xpath.find('local/rounds') > -1:
                        if xpath.find('authentication-method') > -1:
                            rounds['local'][int(round_id) - 1]['authentication-method'] = change['new_value']
                        if xpath.find('secret') > -1:
                            rounds['local'][int(round_id) - 1]['secret'] = change['new_value']
                        if xpath.find('username') > -1:
                            rounds['local'][int(round_id) - 1]['username'] = change['new_value']
                        if xpath.find('password') > -1:
                            rounds['local'][int(round_id) - 1]['password'] = change['new_value']
                        if xpath.find('ca-cert') > -1:
                            rounds['local'][int(round_id) - 1]['ca-cert'] = change['new_value']
                        if xpath.find('cert') > -1:
                            rounds['local'][int(round_id) - 1]['cert'] = change['new_value']
                        if xpath.find('key-pass') > -1:
                            rounds['local'][int(round_id) - 1]['key-pass'] = change['new_value']
                        if xpath.find('key') > -1 and not xpath.find('key-pass') > -1:
                            rounds['local'][int(round_id) - 1]['key'] = change['new_value']
                        if change['new_value']:
                            rounds['local'][int(round_id) - 1]['modified'] = True
                        else:
                            rounds['local'][int(round_id) - 1]['id'] = int(round_id)
                    if xpath.find('remote/rounds') > -1:
                        if xpath.find('authentication-method') > -1:
                            rounds['remote'][int(round_id) - 1]['authentication-method'] = change['new_value']
                        if xpath.find('secret') > -1:
                            rounds['remote'][int(round_id) - 1]['secret'] = change['new_value']
                        if xpath.find('ca-cert') > -1:
                            rounds['remote'][int(round_id) - 1]['ca-cert'] = change['new_value']
                        if xpath.find('cert') > -1:
                            rounds['remote'][int(round_id) - 1]['cert'] = change['new_value']
                        if xpath.find('key') > -1:
                            rounds['remote'][int(round_id) - 1]['key'] = change['new_value']
                        if xpath.find('key-pass') > -1:
                            rounds['remote'][int(round_id) - 1]['key-pass'] = change['new_value']
                        if change['new_value']:
                            rounds['remote'][int(round_id) - 1]['modified'] = True
                        else:
                            rounds['remote'][int(round_id) - 1]['id'] = int(round_id)
                    if xpath.find('local-ts') > -1:
                        local_ts = IPNetwork(change['new_value'])
                        if local_ts == 'None' or local_ts == 'none':
                            del ipsvpn.local_ts
                        else:
                            ipsvpn.local_ts = local_ts
                    if xpath.find('remote-ts') > -1:
                        remote_ts = IPNetwork(change['new_value'])
                        if remote_ts == 'None' or remote_ts == 'none':
                            del ipsvpn.remote_ts
                        else:
                            ipsvpn.remote_ts = remote_ts
                    if xpath.find('enable') > -1:
                        if_status = change['new_value']
                        if if_status == 'true':
                            if_status = True
                        elif if_status == 'false':
                            if_status = False
                    if xpath.find('install-virtual-ip') > -1:
                        inst_vip = False if (change['new_value'] is False or change['new_value'] == 'false') else True
                    if xpath == 'virtual-ip':
                        vip = change['new_value']

                for key, value in old_props.items():
                    for prop in value:
                        if 'enc' in prop and 'auth' in prop:
                            self.logger.debug('Delete proposal: {}'.format(prop))
                            ipsvpn.proposal_del(prop['enc'], prop['auth'], ips.IKE_PHASE.PHASE1 if key == 'ike-phase-1' else ips.IKE_PHASE.PHASE2)

                for key, value in props.items():
                    # phase = ips.IKE_PHASE.PHASE1 if key == 'ike-phase-1' else ips.IKE_PHASE.PHASE2
                    for prop in value:
                        if 'enc' not in prop and 'auth' not in prop:
                            continue
                        elif 'enc' not in prop:
                            prop['enc'] = self.get_proposal_param(key, name, prop['id'], 'encryption-algorithm')
                            ipsvpn.proposal_del(prop['enc'], prop['auth'], ips.IKE_PHASE.PHASE1 if key == 'ike-phase-1' else ips.IKE_PHASE.PHASE2)
                        elif 'auth' not in prop:
                            prop['auth'] = self.get_proposal_param(key, name, prop['id'], 'authentication-algorithm')
                            ipsvpn.proposal_del(prop['enc'], prop['auth'], ips.IKE_PHASE.PHASE1 if key == 'ike-phase-1' else ips.IKE_PHASE.PHASE2)

                        self.logger.debug('Add proposal {}'.format(prop))
                        ipsvpn.proposal_add(prop['enc'], prop['auth'], ips.IKE_PHASE.PHASE1 if key == 'ike-phase-1' else ips.IKE_PHASE.PHASE2)

                # Set local and remote rounds
                for key, value in rounds.items():
                    if key == 'local':
                        for local_round in rounds[key]:
                            if not local_round or 'modified' not in local_round or local_round['modified'] is not True:
                                continue
                            if local_round['authentication-method'] == 'pre-shared-keys':
                                ipsvpn.auth_add(ips.AUTH_SIDE.LOCAL, int(local_round['id']), ips.IPSecAuth(method=ips.AUTH_METHOD.PSK, psk=local_round['secret']))
                            elif local_round['authentication-method'] == 'pubkeys':
                                passphrase = None if 'key-pass' not in local_round.keys() else str(local_round['key-pass'])
                                ipsvpn.auth_add(ips.AUTH_SIDE.LOCAL, int(local_round['id']), ips.IPSecAuth(method=ips.AUTH_METHOD.PUBKEY, ca_crt=local_round['ca-cert'], crt=local_round['cert'], key=local_round['key'], passphrase=passphrase))
                            elif local_round['authentication-method'] == 'xauth':
                                ipsvpn.auth_add(ips.AUTH_SIDE.LOCAL, int(local_round['id']), ips.IPSecAuth(method=ips.AUTH_METHOD.XAUTH, user=local_round['username'], psw=local_round['password']))
                            elif local_round['authentication-method'] == 'eap':
                                passphrase = None if 'key-pass' not in local_round.keys() else str(local_round['key-pass'])
                                ipsvpn.auth_add(ips.AUTH_SIDE.LOCAL, int(local_round['id']), ips.IPSecAuth(method=ips.AUTH_METHOD.EAP, user=local_round['username'], psw=local_round['password'], ca_crt=local_round['ca-cert'], crt=local_round['cert'], key=local_round['key'], passphrase=passphrase))
                    elif key == 'remote':
                        for remote_round in rounds[key]:
                            if not remote_round or 'modified' not in remote_round or remote_round['modified'] is not True:
                                continue
                            if remote_round['authentication-method'] == 'pre-shared-keys':
                                ipsvpn.auth_add(ips.AUTH_SIDE.REMOTE, int(remote_round['id']), ips.IPSecAuth(method=ips.AUTH_METHOD.PSK, psk=remote_round['secret']))
                            elif remote_round['authentication-method'] == 'pubkeys':
                                ipsvpn.auth_add(ips.AUTH_SIDE.REMOTE, int(remote_round['id']), ips.IPSecAuth(method=ips.AUTH_METHOD.PUBKEY, ca_crt=remote_round['ca-cert'], crt=remote_round['cert']))

                if inst_vip is False:
                    ipsvpn.vips = None
                else:
                    if vip is not None:
                        ipsvpn.vips = IPAddress(vip)

                self.__set_ippass_ipsec(prev_status, if_status)

                if prev_status is True:
                    self.__ipsecvpn_disconnect(ipsvpn)
                if if_status is True:
                    self.__ipsecvpn_connect(ipsvpn)

    def __ipsecvpn_for_reenable(self, ifname):
        reenable_ipsecvpn_list = []

        # Proceed only if there is any ipsec vpn
        if len(self.ipsvpn_interfaces) > 0:

            # Get addresses of the adapter, specified by name "ifname"
            adapter_ipv4 = None
            try:
                adapter_ipv4 = self.execute_shell_command("ip addr show {}".format(ifname)).split("inet ")[1].split("/")[0]
            except:
                # On interfaces without inet address exception will be raised, so silently ignore it.
                pass

            adapter_ipv6 = None
            adapter_ipv6_temp = None
            try:
                adapter_ipv6_temp = self.execute_shell_command("ip addr show {} | grep inet6 | grep global ".format(ifname))
                if adapter_ipv6_temp is not None and adapter_ipv6_temp != "":
                    adapter_ipv6 = adapter_ipv6_temp.split("inet6 ")[1].split("/")[0].strip('\n').strip()
                    if adapter_ipv6.count(':') < 2:
                        adapter_ipv6 = None
            except:
                # On interfaces without inet6 address exception will be raised, so silently ignore it.
                pass

            # If the adapter has at least v4 or v6 address proceed
            if adapter_ipv4 is not None or adapter_ipv6 is not None:
                try:
                    adapter_ipv4 = str(adapter_ipv4).encode()
                    adapter_ipv6 = str(adapter_ipv6).encode()
                except:
                    self.logger.error("Error occurred while encoding IP address for IPSec.")

                # Iterate all ipsec vpn interfaces and search for duplicating IP with that of the adapter, specified by ifname
                for ipsecvpn_iface in self.ipsvpn_interfaces:
                    sa = ipsecvpn_iface.get_sa()
                    if sa is not None:
                        (ipsecvpn_iface_ip, ipsecvpn_iface_port) = sa['local'].rsplit(':',1)
                        self.logger.debug("adapter_ipv4={}, adapter_ipv6={}, ipsecvpn_iface_ip={}".format(adapter_ipv4, adapter_ipv6, ipsecvpn_iface_ip))
                        if str(adapter_ipv4) == str(ipsecvpn_iface_ip) or str(adapter_ipv6) == str(ipsecvpn_iface_ip):
                            reenable_ipsecvpn_list.append(ipsecvpn_iface)

        return reenable_ipsecvpn_list

    def _shell_return_ok(self, cmd):
        try:
            self.execute_shell_command(cmd)
        except Exception as ex:
            self.logger.error(ex)
            return False
        return True

    def __set_pci_whitelist(self, action, devname, physical_network, current_list):
        import ast
        import json
        try:
            current_list = ast.literal_eval(str(current_list).strip())
        except Exception as ex:
            self.logger.error("Error parsing PCI current list '{}'".format(current_list))
            self.logger.error(ex)
            current_list = []
        if action == 'add':
            new_pci = {"devname": devname, "physical_network": physical_network}
            current_list.append(new_pci)
        elif action == 'remove':
            current_list[:] = [pci for pci in current_list if pci['devname'] != devname]
        return json.dumps(current_list)

    def init_pci_data(self):
        self.pciids = dict()
        self.igb_vfs = dict()
        self.initramfs_changed = False
        self.igb_recalc_needed = False
        lines = ""

        if os.path.exists("/etc/udev/rules.d/70-persistent-net.rules"):
            with open("/etc/udev/rules.d/70-persistent-net.rules", "r") as f:
                lines = f.readlines()

        for line in lines:
            if 0 == len(line.strip()):
                continue
            line = line.replace('==', '=')
            line = line.replace(' ', '')
            rules = line.strip().split(',')
            rules_dict = dict()
            for rule in rules:
                (key, val) = tuple(rule.split('='))
                rules_dict[key] = val.replace('\"', '')

            if 'net' == rules_dict.get('SUBSYSTEM', 'none') and 'none' != rules_dict.get('NAME', 'none'):
                name = rules_dict.get('NAME', 'none')
                pci = rules_dict.get('KERNELS', 'none')
                tempdict = {name: pci}
                tempdict.update(self.pciids)
                self.pciids = tempdict

        for name, pci in self.pciids.items():
            if os.path.exists("/sys/bus/pci/drivers/igb/%s" % pci):
                self.igb_vfs[name] = 0
            else:
                if 0 == os.system('/sbin/dpdk-devbind --status-dev net | grep "unused=.*igb.*$" | grep "%s" -q' % pci):
                    self.igb_vfs[name] = 0

    def __set_igb_param(self, name, mode):
        if name not in self.igb_vfs:
            return
        if mode == 'sr-iov':
            self.igb_vfs[name] = 7
        else:
            self.igb_vfs[name] = 0

        self.igb_recalc_needed = True

    def _recalc_igb_params(self):
        self.igb_recalc_needed = False
        max_vfs = None
        only_zeroes = True
        for n, vfs in self.igb_vfs.items():
            if max_vfs:
                max_vfs = "%s,%s" % (max_vfs, vfs)
            else:
                max_vfs = "%s" % (vfs)
            if vfs > 0:
                only_zeroes = False

        max_vfs_opt = "options igb max_vfs=%s" % max_vfs
        need_update = False
        try:
            optfile = open("/etc/modprobe.d/igb.conf", "r")
            cur_opt = optfile.read()
            optfile.close()
            if max_vfs_opt != cur_opt:
                if only_zeroes and bool(re.match('^options igb max_vfs=[0,]+$', cur_opt)):
                    need_update = False
                else:
                    need_update = True
        except:
            need_update = True
        try:
            if need_update:
                optfile = open("/etc/modprobe.d/igb.conf", 'w')
                optfile.write("options igb max_vfs=%s" % max_vfs)
                optfile.close()
                os.system('update-initramfs -u')
                self.initramfs_changed = True
        except:
            self.logger.error("Couldn't update \"/etc/modprobe.d/igb.conf\" file.")

    def ippass_add_del_management_entries(self, ifname, ip_addr, add):
        if_idx = None
        if_name = None
        if not ip_addr or not ifname:
            return

        vpp_if = vppif.VppApiInterface('ippass_add_del_management_entries')
        if not vpp_if:
            return
            
        (if_name, if_idx) = vpp_if.search_vppif_by_hostifname(ifname)
        if if_name == None or if_idx == None:
            del(vpp_if)
            return

        ifname = 'host-' + ifname if self.get_interface_param(ifname, 'mode') == 'virtio' else ifname
        ifname = self.tag_interface(ifname, self.get_interface_param(ifname.replace('host-', ''), 'vlan'))
        if_idx = vpp_if.get_ifidx_by_name(ifname)
        self.logger.info('IPPass management: ifname {}, ip_addr {}, if_idx {}, add {}' .format(ifname, ip_addr, if_idx, add))
        host_idx = vpp_if.ippass_get_host_if_idx(if_idx)
        if host_idx is None:
            del(vpp_if)
            return

        ret = vpp_if.ippass_check_status(host_idx=host_idx)
        if ret.retval != 0:
            del(vpp_if)
            return

        # add / del CLI entry:
        if 0 != vpp_if.add_del_ippass_permit_host_entry(host_idx=host_idx, dir_out=0, protocol='tcp', dst_port=CLI_PORT, ip=ip_addr, add=add):
            self.logger.error('IPPass - failed to apply management rules')
        else:
            self.logger.info('IPPass - {} permit CLI entry' .format('Added' if add else 'Deleted'))

        # add / det TVEAPI entry:
        if 0 != vpp_if.add_del_ippass_permit_host_entry(host_idx=host_idx, dir_out=0, protocol='tcp', dst_port=TVEAPI_PORT, ip=ip_addr, add=add):
            self.logger.error('IPPass - failed to apply management rules')
        else:
            self.logger.info('IPPass - {} permit TVE API entry' .format('Added' if add else 'Deleted'))

        # add / det Keystone auth ports entries:
        if 0 != vpp_if.add_del_ippass_permit_host_entry(host_idx=host_idx, dir_out=0, protocol='tcp', dst_port=5000, ip=ip_addr, add=add):
            self.logger.error('IPPass - failed to apply management rules')
        else:
            self.logger.info('IPPass - {} permit port 5000 entry' .format('Added' if add else 'Deleted'))

        # add / det Keystone auth ports entries:
        if 0 != vpp_if.add_del_ippass_permit_host_entry(host_idx=host_idx, dir_out=0, protocol='tcp', dst_port=35357, ip=ip_addr, add=add):
            self.logger.error('IPPass - failed to apply management rules')
        else:
            self.logger.info('IPPass - {} permit port 35357 entry' .format('Added' if add else 'Deleted'))

        del(vpp_if)

    def __is_svc_masked(self, svc):
        try:
            state = subprocess.check_output(['systemctl', 'is-enabled', svc])
            if state.decode().rstrip() == 'masked':
                return True
        except subprocess.CalledProcessError as e:
            if e.output.decode().rstrip() == 'masked':
                return True

        return False

    def __set_ippass_management(self, name, b_set, b_old, b_new, m_set, m_old, m_new, ip_set, ip_old, ip_new, ip6_set, ip6_old, ip6_new):
        to_del = False
        to_add = False
        apply_addr = ip_new
        apply_addr6 = ip6_new

        if b_set == False and m_set == False and ip_set == False and ip6_set == False:
            return

        # Check if old entry should be deleted from vpp
        if b_set == True or m_set == True:
            if b_new in ['dhcp', 'dhcp6', 'dhcp46'] or m_new == False or m_new == 'false':
                to_del = True
                if ip_set == False or (ip_set == True and ip_new == None):
                    apply_addr = ip_old
                if ip6_set == False or (ip6_set == True and ip6_new == None):
                    apply_addr6 = ip6_old

        if ip_set == True:
            if ip_new == None or (ip_new != None and ip_new != ip_old):
                to_del = True
                apply_addr = ip_old

        if ip6_set == True:
            if ip6_new == None or (ip6_new != None and ip6_new != ip6_old):
                to_del = True
                apply_addr6 = ip6_old

        # Delete entry
        if to_del == True and apply_addr != None:
            ip = IPNetwork(apply_addr)
            self.logger.info('IPPass - delete old management entry for interface {} address {} '.format(name, str(ip.ip)))
            self.ippass_add_del_management_entries(name, str(ip.ip), 0)

        if to_del == True and apply_addr6 != None:
            ip = IPNetwork(apply_addr6)
            self.logger.info('IPPass - delete old management entry for interface {} address {} '.format(name, str(ip.ip)))
            self.ippass_add_del_management_entries(name, str(ip.ip), 0)

        # Check if new entry should be installed
        if b_set == True or m_set == True:
            if (b_new != None and b_new not in ['dhcp', 'dhcp6', 'dhcp46']) or m_new == True or m_new == 'true':
                to_add = True
                apply_addr = ip_new
                if ip_set == False:
                    apply_addr = ip_old

                apply_addr6 = ip6_new
                if ip6_set == False:
                    apply_addr6 = ip6_old

        else:
            if ip_set == True and ip_new != None:
                if (m_old == True or m_old == 'true') and (b_old != None and b_old not in ['dhcp', 'dhcp6', 'dhcp46']):
                    to_add = True
                    apply_addr = ip_new
            if ip6_set == True and ip6_new != None:
                if (m_old == True or m_old == 'true') and (b_old != None and b_old not in ['dhcp', 'dhcp6', 'dhcp46']):
                    to_add = True
                    apply_addr6 = ip6_new

        # Add entry
        if to_add == True:
            if ip_new == None:
                apply_addr = ip_old
            if apply_addr != None:
                ip = IPNetwork(apply_addr)
                self.logger.info('IPPass - add new entry management for interface {} address {} '.format(name, str(ip.ip)))
                self.ippass_add_del_management_entries(name, str(ip.ip), 1)
            if ip6_new == None:
                apply_addr6 = ip6_old
            if apply_addr6 != None:
                ip = IPNetwork(apply_addr6)
                self.logger.info('IPPass - add new entry management for interface {} address {} '.format(name, str(ip.ip)))
                self.ippass_add_del_management_entries(name, str(ip.ip), 1)

    def __config_interface_ippass(self, name, config, change_set):
        ippass_en = config.get('ip-passthrough', False)

        if False == ippass_en or self.vpp_not_running():
            return

        self.logger.info("On interface {} ippass is enabled".format(name))

        # if config and 'enable' in config:
        #     old_enable_nat = config['enable']

        vpp_if = vppif.VppApiInterface('sr-interfaces')
        vpp_if_id = None
        host_if_id = None

        if change_set:

            if config.get('mode', 'virtio') == 'dpdk':
                vpp_if_id = vpp_if.get_ifidx_by_name(name)
            else:
                vpp_if_id = vpp_if.get_ifidx_by_name("host-{}".format(name))

            if vpp_if_id is None:
                del(vpp_if)
                return

            # Below code must be on top because of prior vlan handling logic
            for xpath, change in change_set.items():
                if xpath.find('vlan') > -1:
                    vpp_if_name = vpp_if.get_ifname_by_ifidx(vpp_if_id)
                    vlan_id = change.get('new_value', None) if change.get('operation', None) != 'DELETED' else None
                    vpp_if_name = self.tag_interface(vpp_if_name, vlan_id)
                    vpp_if_id = vpp_if.get_ifidx_by_name(vpp_if_name)
                    if vpp_if_id is None:
                        del(vpp_if)
                        return
                    host_if_id = vpp_if.ippass_get_host_if_idx(vpp_if_id)
                    if host_if_id is None:
                        del(vpp_if)
                        return
                    if config.get('bootproto', '') in ['dhcp', 'dhcp6', 'dhcp46']:
                        self.logger.info('Enable dhcp on ippass host inteface')
                        vpp_if.add_del_ippass_permit_host_entry(host_if_id, 'dhcp')
                else:
                    vpp_if_name = vpp_if.get_ifname_by_ifidx(vpp_if_id)
                    vpp_if_name = self.tag_interface(vpp_if_name, self.get_interface_param(name, 'vlan'))
                    vpp_if_id = vpp_if.get_ifidx_by_name(vpp_if_name)
                    host_if_id = vpp_if.ippass_get_host_if_idx(vpp_if_id)
                    if host_if_id is None:
                        del(vpp_if)
                        return

            b_set = False
            bproto = None
            b_old = self.get_interface_param(name, 'bootproto')

            m_set = False
            management = None
            m_old = self.get_interface_param(name, 'management')

            ipv4_set = False
            ipv4_addr_mask = None
            ipv4_old = self.get_interface_param(name, 'ipv4-prefix')

            ipv6_set = False
            ipv6_addr_mask = None
            ipv6_old = self.get_interface_param(name, 'ipv6-prefix')

            for xpath, change in change_set.items():
                if xpath.find('bootproto') > -1:
                    b_set = True
                    bproto = change.get("new_value", None)
                    if change.get("old_value","") in ['dhcp', 'dhcp6', 'dhcp46'] and change.get("new_value","") not in ['dhcp', 'dhcp6', 'dhcp46']:
                        vpp_if.add_del_ippass_permit_host_entry(host_if_id, 'dhcp', add=0)

                    if change.get("new_value","") in ['dhcp', 'dhcp6', 'dhcp46'] and change.get("old_value","") not in ['dhcp', 'dhcp6', 'dhcp46']:
                        vpp_if.add_del_ippass_permit_host_entry(host_if_id, 'dhcp')
                if xpath.find('management') > -1:
                    m_set = True
                    management = change.get("new_value", None)
                if xpath.find('ipv4-prefix') > -1:
                    ipv4_set = True
                    if change.get('operation', None) != 'DELETED':
                        ipv4_addr_mask = change.get("new_value", None)
                if xpath.find('ipv6-prefix') > -1:
                    ipv6_set = True
                    if change.get('operation', None) != 'DELETED':
                        ipv6_addr_mask = change.get("new_value", None)

            del(vpp_if)
            self.__set_ippass_management(name, b_set, b_old, bproto, m_set, m_old, management,
                                         ipv4_set, ipv4_old, ipv4_addr_mask, ipv6_set, ipv6_old, ipv6_addr_mask)

        else:
            self.logger.info("IPPass apply startup config")
            name = self.tag_interface(name, self.get_interface_param(name, 'vlan'))
            if config.get('mode', 'virtio') == 'dpdk':
                vpp_if_id = vpp_if.get_ifidx_by_name(name)
            else:
                vpp_if_id = vpp_if.get_ifidx_by_name("host-{}".format(name))

            if vpp_if_id is None:
                del(vpp_if)
                return

            host_if_id = vpp_if.ippass_get_host_if_idx(vpp_if_id)
            if host_if_id is None:
                del(vpp_if)
                return

            name = name.split('.')[0]

            if config.get('bootproto', '') in ['dhcp', 'dhcp6', 'dhcp46']:
                self.logger.info('Enable dhcp on ippass host inteface')
                vpp_if.add_del_ippass_permit_host_entry(host_if_id, 'dhcp')

            del(vpp_if)
            if config.get('management', False) == True or config.get('management', False) == 'true':
                if config.get('bootproto', '') == 'none':
                    ipv4_addr_mask = config.get('ipv4-prefix', '')
                    if ipv4_addr_mask != '':
                        ip = IPNetwork(ipv4_addr_mask)
                        self.ippass_add_del_management_entries(name, str(ip.ip), 1)

                    ipv6_addr_mask = config.get('ipv6-prefix', '')
                    if ipv6_addr_mask != '':
                        ip = IPNetwork(ipv6_addr_mask)
                        self.ippass_add_del_management_entries(name, str(ip.ip), 1)

    def __config_interface_loopback(self, name, config, change_set):
        self.logger.debug("__config_interface_loopback")
        self.logger.debug("Config {}".format(config))
        self.logger.debug("Change set {}".format(change_set))

        # Getting network-mode status
        mode = "VPP" if self.get_network_mode() == 'vpp' else "OVS"
        self.logger.debug(f"Network mode: {mode}")

        if not change_set and config:
            # Startup config
            loop_if = next((a for a in self.loopback_interfaces if a.name == config.get('name')), None)
            if loop_if:
                self.logger.debug("Interface \'{}\' already exists.".format(loop_if.name))
            else:
                ipv4 = config.get('ipv4-prefix', None)
                if ipv4:
                    ipv4 = ipaddress.ip_interface(ipv4)

                ipv6 = config.get('ipv6-prefix', None)
                if ipv6:
                    ipv6 = ipaddress.ip_interface(ipv6)

                enable = config.get('enable', False)

                loop_if = loopif.LoopInterface (config.get('name'), ipv4_prefix=ipv4, ipv6_prefix=ipv6, enable=enable, logger=self.logger, network_mode=mode)
                self.loopback_interfaces.append(loop_if)
            return

        if change_set:
            ch_set_keys = change_set.keys()

            self.logger.info('Keys : {}'.format(ch_set_keys))
            loop_if = next((loop for loop in self.loopback_interfaces if loop.name == name), None)
            if 'name' in ch_set_keys:
                if change_set.get('name', None):
                    if change_set.get('name').get('operation', None) == 'CREATED':
                        if loop_if:
                            self.logger.error("There is object for {} ?!?!".format(name))
                        else:
                            loop_if = loopif.LoopInterface(name, logger=self.logger, network_mode=mode)
                            self.loopback_interfaces.append(loop_if)
                    elif change_set.get('name').get('operation', None) == 'DELETED':
                        if loop_if:
                            self.loopback_interfaces.remove(loop_if)
                            del loop_if
                        return

            if 'enable' in ch_set_keys:
                oper = change_set.get('enable').get('operation', None)
                if oper == 'MODIFIED' or oper == 'CREATED':
                    enable = change_set.get('enable').get('new_value')
                    enable = True if enable == 'true' else False
                else:
                    enable = False

                self.logger.debug(str(loop_if))
                loop_if.enable = enable

            if 'ipv4-prefix' in ch_set_keys:
                oper = change_set.get('ipv4-prefix').get('operation', None)
                if oper == 'MODIFIED' or oper == 'CREATED':
                    ipv4_pr = change_set.get('ipv4-prefix').get('new_value')
                    loop_if.ip4 = ipaddress.IPv4Interface(str(ipv4_pr))
                else:
                    loop_if.ip4 = None
                loop_if.enable = False
                loop_if.enable = True


            if 'ipv6-prefix' in ch_set_keys:
                oper = change_set.get('ipv6-prefix').get('operation', None)
                if oper == 'MODIFIED' or oper == 'CREATED':
                    ipv6_pr = change_set.get('ipv6-prefix').get('new_value')
                    loop_if.ip6 = ipaddress.IPv6Interface(str(ipv6_pr))
                else:
                    loop_if.ip6 = None
                loop_if.enable = False
                loop_if.enable = True

            for key in change_set.keys():
                self.logger.debug('Key \'{}\' operation {}'.format(key, change_set.get(key).get('operation')))

    def __config_interface_vxlan(self, name, config, change_set):
        if self.get_network_mode() != 'vpp':
            self.logger.error('uCPE is not in VPP mode')
            return
        self.logger.debug("__config_interface_vxlan")
        self.logger.debug("Config {}".format(config))
        self.logger.debug("Change set {}".format(change_set))

        local = None
        lvtepv4 = None
        lvtepv6 = None
        old_local = None
        old_localv6 = None
        old_remote = []
        old_type = None
        lvtep = None
        rvteps_add = []
        rvteps_del = []
        tun_type = None
        mcast = None
        mcast_ip = None
        old_mcast_ip = None

        if config:
            if config.get('local', {}).get('vtep-addr', None) != None:
                local = config.get('local', {}).get('vtep-addr', None)
                if local:
                    old_local = ipaddress.ip_address(local) if isinstance(ipaddress.ip_address(local), ipaddress.IPv4Address) else None
                    old_localv6 = ipaddress.ip_address(local) if isinstance(ipaddress.ip_address(local), ipaddress.IPv6Address) else None
            elif config.get('local', {}).get('vtep-if', None) != None:
                local = config.get('local').get('vtep-if', None)
                if local != None:
                    bootproto = None
                    address = None
                    addressv6 = None
                    bootproto = self.get_item("/ucpe:config/interfaces:interface[name='{}']/bootproto".format(local))
                    if bootproto == "none" and bootproto != None:
                        addressv6 = self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv6-prefix".format(local))
                        if addressv6 is not None:
                            old_localv6 = ipaddress.ip_interface(addressv6).ip

                        address = self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv4-prefix".format(local))
                        if address is not None:
                            old_local = ipaddress.ip_interface(address).ip

            for rvtep_addr in config.get('remote', {}).get('vtep-addrs', []):
                old_remote.append(ipaddress.ip_address(rvtep_addr.get('vtep-addr')))

            old_type = config.get('tunnel-type', 'vxlan')

            if config.get('remote', None) != None:
                mcast = config.get('remote').get('mcast-vtep-addr', None)
                if mcast != None:
                    old_mcast_ip = ipaddress.ip_address(mcast)

        if not change_set and config:
            # Startup config
            vxlan_if = next((a for a in self.vxlan_interfaces if a.name == config.get('name')), None)
            tun_type = vxlif.TUNNEL_TYPE.VXLAN
            if vxlan_if:
                self.logger.debug("Interface \'{}\' allready exists.".format(vxlan_if.name))
            else:
                vni = config.get('vni', None)
                shg = config.get('shg', None)

                if old_type == 'vxlan-gpe':
                    tun_type = vxlif.TUNNEL_TYPE.VXLAN_GPE

                vxlan_if = vxlif.VxLanInterface( config.get('name'), vni=vni, lvtep=old_local, lvtepv6=old_localv6, rvteps=old_remote, tun_type=tun_type,
                                                 logger=self.logger, shg=shg, mcast_if=local, mcast_ip=old_mcast_ip)
                self.vxlan_interfaces.append(vxlan_if)
                enable = config.get('enable', False)
                vxlan_if.enable = enable

                ipv4_pr = config.get('ipv4-prefix', None)
                if ipv4_pr:
                    vxlan_if.ip4 = ipaddress.IPv4Interface(str(ipv4_pr))

                ipv6_pr = config.get('ipv6-prefix', None)
                if ipv6_pr:
                    vxlan_if.ip6 = ipaddress.IPv6Interface(str(ipv6_pr))

                bootproto = config.get('bootproto', None)
                bp_set = None
                if bootproto:
                    if bootproto == 'none':
                        bp_set = vxlif.BOOT_PROTO.NONE
                    elif bootproto == 'dhcp':
                        bp_set = vxlif.BOOT_PROTO.DHCP
                    elif bootproto == 'dhcp6':
                        bp_set = vxlif.BOOT_PROTO.DHCP6
                    elif bootproto == 'dhcp46':
                        bp_set = vxlif.BOOT_PROTO.DHCP46

                    if bp_set:
                        vxlan_if.bootproto = bp_set
                dns = config.get('dns', None)
                if dns:
                    self.logger.debug('! Setting DNS with name {}'.format(name))
                    dnses = dns.split(',')
                    dnsopts = ''.join(map(lambda dns_srv: ' '+ dns_srv, dnses))
                    subprocess.call('/usr/bin/resolvectl dns br-{} {}'.format(name, dnsopts), shell=True)

            if old_local:
                self.__config_ippass_vxlan(None, [], None, old_local, old_remote, old_type)
            if old_localv6:
                self.__config_ippass_vxlan(None, [], None, old_localv6, old_remote, old_type)
            return

        if change_set:
            ch_set_keys = change_set.keys()
            lvtep_set = False
            rvteps_set = False
            tuntype_set = False

            self.logger.info('Keys : {}'.format(ch_set_keys))
            vxlan_if = next((vxlan for vxlan in self.vxlan_interfaces if vxlan.name == name), None)

            if 'name' in ch_set_keys:
                if change_set.get('name', None):
                    if change_set.get('name').get('operation', None) == 'CREATED':
                        if vxlan_if:
                            self.logger.error("There is object for {} ?!?!".format(name))
                        else:
                            vxlan_if = vxlif.VxLanInterface(name, logger=self.logger)
                            self.vxlan_interfaces.append(vxlan_if)
                    elif change_set.get('name').get('operation', None) == 'DELETED':
                        if vxlan_if:
                            self.vxlan_interfaces.remove(vxlan_if)
                            del vxlan_if
                        if old_local:
                            self.__config_ippass_vxlan(old_local, old_remote, old_type, None, [], None)
                        if old_localv6:
                            self.__config_ippass_vxlan(old_localv6, old_remote, old_type, None, [], None)
                        return

            if 'vni' in ch_set_keys:
                self.logger.debug('VNI is changed')

                oper = change_set.get('vni').get('operation', None)
                vni = None

                if oper == 'MODIFIED' or oper == 'CREATED':
                    vni = change_set.get('vni').get('new_value', None)

                vxlan_if.vni = int(vni)

                self.logger.info('VNI operation is {}'.format(oper))

            if 'shg' in ch_set_keys:
                self.logger.debug('SHG is changed')

                oper = change_set.get('shg').get('operation', None)
                shg = 100

                if oper == 'MODIFIED' or oper == 'CREATED':
                    shg = change_set.get('shg').get('new_value', None)

                vxlan_if.shg = int(shg)

                self.logger.info('SHG operation is {}'.format(oper))

            if 'local/vtep-addr' in ch_set_keys:
                self.logger.debug('local/vtep-addr is changed')
                oper = change_set.get('local/vtep-addr').get('operation', None)
                lvtep_set = True
                lvtep = None
                if oper == 'CREATED' or oper == 'MODIFIED':
                    lvtep = change_set.get('local/vtep-addr').get('new_value', None)
                elif oper == 'DELETED':
                    if 'local/vtep-if' not in ch_set_keys:
                        self.logger.debug('local/vtep-addr is deleted')
                        vxlan_if.lvtep = None

                if lvtep:
                    lvtep = ipaddress.ip_address(lvtep)
                    if lvtep.version == 6:
                        if old_local:
                            vxlan_if.lvtep = None

                        vxlan_if.lvtepv6 = lvtep
                        lvtepv6 = lvtep
                    else:
                        if old_localv6:
                            vxlan_if.lvtepv6 = None

                        vxlan_if.lvtep = lvtep
                        lvtepv4 = lvtep

                self.logger.debug('LOCAL ip operation is {}'.format(oper))

            if 'local/vtep-if' in ch_set_keys:
                self.logger.debug('local/vtep-if is changed')
                oper = change_set.get('local/vtep-if').get('operation', None)
                lvtep_set = True
                lvtep = None
                if oper == 'CREATED' or oper == 'MODIFIED':
                    lvtep = change_set.get('local/vtep-if').get('new_value', None)
                    vxlan_if.mcast_if = lvtep
                elif oper == 'DELETED':
                    if 'local/vtep-addr' not in ch_set_keys:
                        self.logger.debug('local/vtep-if is deleted')
                        vxlan_if.lvtep = None
                        vxlan_if.lvtepv6 = None

                    vxlan_if.mcast_if = None

                if lvtep:
                    bootproto = None
                    address = None
                    addressv6 = None
                    bootproto = self.get_item("/ucpe:config/interfaces:interface[name='{}']/bootproto".format(lvtep))
                    if (bootproto == "none" and bootproto != None) or "lo" in lvtep:
                        addressv6=self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv6-prefix".format(lvtep))
                        if addressv6 is not None:
                            lvtepv6 = ipaddress.ip_interface(addressv6).ip

                        address = self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv4-prefix".format(lvtep))
                        if address is None:
                            address = change_set.get('local/lvtep-if-ipv4-prefix', None).get('new_value', None)
                        if address is not None:
                            lvtepv4 = ipaddress.ip_interface(address).ip
                    else:
                        lvtepv4 = None

                    if lvtepv4:
                        vxlan_if.lvtep = lvtepv4
                    if lvtepv6:
                        vxlan_if.lvtepv6 = lvtepv6

                self.logger.debug('LOCAL if operation is {}'.format(oper))


            p = re.compile("^remote\/vtep-addrs\[vtep-addr='.*'\]\/vtep-addr")

            for key in ch_set_keys:
                if p.match(key):
                    oper = change_set.get(key).get('operation', None)
                    rvteps_set = True
                    if oper == 'CREATED':
                        rvteps_add.append(ipaddress.ip_address(change_set.get(key).get('new_value')))
                    elif oper == 'DELETED':
                        rvteps_del.append(ipaddress.ip_address(change_set.get(key).get('old_value')))
                    else:
                        self.logger.info("Unhandled operation for REMOTE {}".format(oper))

                    self.logger.info('REMOTE operation is {}'.format(oper))

            if rvteps_add:
                vxlan_if.add_rvteps(rvteps_add)

            if rvteps_del:
                vxlan_if.del_rvteps(rvteps_del)

            if 'remote/mcast-vtep-addr' in ch_set_keys:
                self.logger.debug('remote/mcast-vtep-addr is changed')
                oper = change_set.get('remote/mcast-vtep-addr').get('operation', None)
                mcast = None
                if oper == 'CREATED' or oper == 'MODIFIED':
                    mcast = change_set.get('remote/mcast-vtep-addr').get('new_value', None)
                    vxlan_if.mcast_ip = ipaddress.ip_address(mcast)
                elif oper == 'DELETED':
                    self.logger.debug('remote/mcast-vtep-addr is deleted')
                    vxlan_if.mcast_ip = None

                # if lvtep:
                #     lvtep = ipaddress.ip_address(lvtep)
                #     vxlan_if.lvtep = lvtep

            if 'enable' in ch_set_keys:
                oper = change_set.get('enable').get('operation', None)
                if oper == 'MODIFIED' or oper == 'CREATED':
                    enable = change_set.get('enable').get('new_value')
                    enable = True if enable == 'true' else False
                else:
                    enable = False

                vxlan_if.enable = enable

            if 'ipv4-prefix' in ch_set_keys:
                oper = change_set.get('ipv4-prefix').get('operation', None)
                if oper == 'MODIFIED' or oper == 'CREATED':
                    ipv4_pr = change_set.get('ipv4-prefix').get('new_value')
                    vxlan_if.ip4 = ipaddress.IPv4Interface(str(ipv4_pr))
                else:
                    vxlan_if.ip4 = None

            if 'ipv6-prefix' in ch_set_keys:
                oper = change_set.get('ipv6-prefix').get('operation', None)
                if oper == 'MODIFIED' or oper == 'CREATED':
                    ipv6_pr = change_set.get('ipv6-prefix').get('new_value')
                    vxlan_if.ip6 = ipaddress.IPv6Interface(str(ipv6_pr))
                else:
                    vxlan_if.ip6 = None

            if 'bootproto' in ch_set_keys:
                oper = change_set.get('bootproto').get('operation', None)
                bp_set = None
                if oper == 'MODIFIED' or oper == 'CREATED':
                    bootproto = change_set.get('bootproto').get('new_value')

                    if bootproto == 'none':
                        bp_set = vxlif.BOOT_PROTO.NONE
                    elif bootproto == 'dhcp':
                        bp_set = vxlif.BOOT_PROTO.DHCP
                    elif bootproto == 'dhcp6':
                        bp_set = vxlif.BOOT_PROTO.DHCP6
                    elif bootproto == 'dhcp46':
                        bp_set = vxlif.BOOT_PROTO.DHCP46

                    if bp_set:
                        vxlan_if.bootproto = bp_set
                else:
                    vxlan_if.bootproto = vxlif.BOOT_PROTO.NONE

            if 'tunnel-type' in ch_set_keys:
                oper = change_set.get('tunnel-type').get('operation', None)
                tuntype_set = True
                type_set = None
                if oper == 'MODIFIED' or oper == 'CREATED':
                    tun_type = change_set.get('tunnel-type').get('new_value')

                    if tun_type == 'vxlan':
                        type_set = vxlif.TUNNEL_TYPE.VXLAN
                    elif tun_type == 'vxlan-gpe':
                        type_set = vxlif.TUNNEL_TYPE.VXLAN_GPE

                    if type_set:
                        vxlan_if.tun_type = type_set
                else:
                    vxlan_if.tun_type = vxlif.TUNNEL_TYPE.VXLAN

            if 'dns' in ch_set_keys:
                oper = change_set.get('dns').get('operation', None)
                if oper == 'MODIFIED' or oper == 'CREATED':
                    dns = change_set.get('dns').get('new_value')
                    dnses = dns.split(',')
                    dnsopts = ''.join(map(lambda dns_srv: ' '+ dns_srv, dnses))                    
                    subprocess.call('/usr/bin/resolvectl dns br-{} {}'.format(name, dnsopts), shell=True)
                else:
                    dns = change_set.get('dns').get('old_value')
                    dnses = dns.split(',')
                    # revert all !?
                    subprocess.call('/usr/bin/resolvectl revert br-{}'.format(name), shell=True)

            for key in change_set.keys():
                self.logger.debug('Key \'{}\' operation {}'.format(key, change_set.get(key).get('operation')))

            if lvtep_set or rvteps_set or tuntype_set:
                if not rvteps_set:
                    rvteps_del = old_remote
                    rvteps_add = old_remote
                if not lvtep_set:
                    lvtepv4 = old_local
                    lvtepv6 = old_localv6
                if not tuntype_set:
                    tun_type = old_type

                self.__config_ippass_vxlan(old_local, rvteps_del, old_type,
                                                lvtepv4, rvteps_add, tun_type, False)
                self.__config_ippass_vxlan(old_localv6, rvteps_del, old_type,
                                                lvtepv6, rvteps_add, tun_type, False)

    def __config_interface_pppoe(self, name, config, change_set):
        if self.get_network_mode() != 'vpp':
            self.logger.error('uCPE is not in VPP mode')
            return

        c_auth_type = pppoeif.AUTH_TYPE.PAP
        c_username = None
        c_password = None
        c_ethif = None
        c_status = False
        c_idle = 0
        c_retry = 0
        c_metric = 130
        c_management = False
        c_ac_name = None
        c_ippass = False

        self.logger.debug("Config {}".format(config))
        if config:
            if config.get('auth-type', 'pap') == 'chap':
                c_auth_type = pppoeif.AUTH_TYPE.CHAP
            
            c_username = config.get('username', None)
            c_password = config.get('password', None)
            c_ethif = config.get('ethernet-interface', None)
            c_status = config.get('enable', False)
            c_idle = config.get('idle-timeout', 0)
            c_retry = config.get('retry-timeout', 0)
            c_metric = config.get('metric', 130)
            c_management = config.get('management', False)
            c_ac_name = config.get('ac-name', None)
            c_ipv6 = config.get('ipv6', False)
            c_ippass = config.get('ip-passthrough', False)

            ifname = self.__get_ifname(name, None)

            #To-Do get ethernet-interface mode
            if c_ethif:
                if 'virtio' == self.get_interface_param(c_ethif, 'mode', 'virtio'):
                    c_ethif = 'host-{}'.format(c_ethif)
                c_ethif = self.tag_interface(c_ethif,  self.get_interface_param(c_ethif.replace('host-', ''), 'vlan', None))

        if not change_set and config:
            # Startup config
            self.logger.debug("APPLYING STARTUP FOR PPPOE INTERFACE - {}".format(name))
            ppp_if = next(( a for a in self.pppoe_interfaces if a.name == config.get('name')), None)
            if ppp_if:
                self.logger.debug("Interface \'{}\' allready exists.".format(ppp_if.name))
            else:
                self.logger.debug("Add PPPoE configuration for \'{}\' interface" .format(config.get('name')))
                ppp_if = pppoeif.PPPoEClientInterface(config.get('name'), auth_type=c_auth_type, username=c_username, password=c_password, 
                                                        ethif=c_ethif, logger=self.logger, enable=c_status, metric=c_metric,
                                                        idle_timeout=c_idle, retry_timeout=c_retry, ipv6=c_ipv6, ac_name=c_ac_name,
                                                        ip_pass=c_ippass)

                self.pppoe_interfaces.append(ppp_if)

                mtu = config.get('mtu', 1492)
                if mtu:
                    ppp_if.mtu = int(mtu)

                self.__apply_management_rules(ifname, False, c_management, change_set)

        if change_set:
            self.logger.debug("APPLYING RUNNING  FOR PPOE INTERFACE - {}".format(name))
            ch_set_keys = change_set.keys()
            self.logger.debug('Keys : {}'.format(ch_set_keys))
            # n_status = c_status

            ppp_if = next(( a for a in self.pppoe_interfaces if a.name == name), None)

            if 'name' in ch_set_keys:
                if change_set.get('name', None):
                    if change_set.get('name').get('operation', None) == 'CREATED':
                        if ppp_if:
                            self.logger.error("There is PPPoE interface object for {} ?!?!".format(name))
                        else:
                            ppp_if = pppoeif.PPPoEClientInterface( name, logger=self.logger )
                            self.pppoe_interfaces.append(ppp_if)
                    elif change_set.get('name').get('operation', None) == 'DELETED':
                        if ppp_if:
                            self.pppoe_interfaces.remove(ppp_if)
                            del ppp_if
                        return

            # # if current status is enable - disable PPPoE connection (in order to apply all new values) and if new status is enable - reconnect at the end
            # if c_status is True:
            #     ppp_if.enable = False

            if 'auth-type' in ch_set_keys:
                oper = change_set.get('auth-type').get('operation', None )
                n_auth_type = None

                self.logger.debug('Auth type is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    if change_set.get('auth-type').get('new_value', 'pap') == 'pap':
                        n_auth_type = pppoeif.AUTH_TYPE.PAP
                    elif change_set.get('auth-type').get('new_value', 'pap') == 'chap':
                        n_auth_type = pppoeif.AUTH_TYPE.CHAP

                ppp_if.auth_type = n_auth_type

            if 'ethernet-interface' in ch_set_keys:
                oper = change_set.get('ethernet-interface').get('operation', None )
                n_ethif = None

                self.logger.debug('Eth if is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_ethif = change_set.get('ethernet-interface').get('new_value', None)

                if n_ethif:
                    if 'virtio' == self.get_interface_param(n_ethif, 'mode', 'virtio'):
                        n_ethif = 'host-{}'.format(n_ethif)
                    if 'eth-if-vlan' in ch_set_keys:
                        vlan = change_set.get('eth-if-vlan').get('new_value')
                        n_ethif += f'.{vlan}'
                    else:
                        n_ethif = self.tag_interface(n_ethif, self.get_interface_param(n_ethif.replace('host-', ''), 'vlan',
                                                                                   None))
                ppp_if.ethif = n_ethif

            if 'username' in ch_set_keys:
                oper = change_set.get('username').get('operation', None )
                n_username = None

                self.logger.debug('Username is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_username = change_set.get('username').get('new_value', None)

                ppp_if.username = n_username

            if 'password' in ch_set_keys:
                oper = change_set.get('password').get('operation', None )
                n_password = None

                self.logger.debug('Password is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_password = change_set.get('password').get('new_value', None)

                ppp_if.password = n_password

            if 'metric' in ch_set_keys:
                oper = change_set.get('metric').get('operation', None )
                n_metric = 130

                self.logger.debug('Metric is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_metric = change_set.get('metric').get('new_value', 130)

                ppp_if.metric = int(n_metric)
                if oper == 'MODIFIED' and c_status is not False:
                    ipdb = IPDB()
                    remote_ip = ''
                    local_ip = ''

                    if name in ipdb.interfaces:
                        ppp_idx = ipdb.interfaces[name]
                        for rt in ipdb.routes:
                            if rt['family'] == 10:
                                continue
                            if rt['oif'] == ppp_idx['index']:
                                if rt['prefsrc']:
                                    local_ip = rt['prefsrc']
                                if rt['dst']:
                                    remote_ip = rt['dst'].split('/')[0]

                    try:
                        cmd = ['/opt/bin/pppif_host_cfg.py', '-n', name, '-l', local_ip, '-r', remote_ip]
                        subprocess.check_call(cmd)
                    except subprocess.CalledProcessError as e:
                        self.logger.debug(e.output)


            if 'idle-timeout' in ch_set_keys:
                oper = change_set.get('idle-timeout').get('operation', None )
                n_idle_timeout = 0

                self.logger.debug('Idle timeout is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_idle_timeout = change_set.get('idle-timeout').get('new_value', 0)

                ppp_if.idle_timeout = int(n_idle_timeout)

            if 'retry-timeout' in ch_set_keys:
                oper = change_set.get('retry-timeout').get('operation', None )
                n_retry_timeout = 0

                self.logger.debug('Retry timeout is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_retry_timeout = change_set.get('retry-timeout').get('new_value', 0)

                ppp_if.retry_timeout = int(n_retry_timeout)

            if 'mtu' in ch_set_keys:
                oper = change_set.get('mtu').get('operation', None )
                n_mtu = 1492

                self.logger.debug('MTU is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_mtu = change_set.get('mtu').get('new_value', 1492)

                ppp_if.mtu = int(n_mtu)

            if 'enable' in ch_set_keys:
                oper = change_set.get('enable').get('operation', None)
                n_status = False

                self.logger.debug('State is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    if change_set.get('enable').get('new_value', False) == 'true' or \
                            change_set.get('enable').get('new_value', False) == 'True' or \
                            change_set.get('enable').get('new_value', False) == True:
                        n_status = True
                    
                ppp_if.enable = n_status

            if 'management' in ch_set_keys:
                oper = change_set.get('management').get('operation', None)
                n_management = False

                self.logger.debug('Management is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    if change_set.get('management').get('new_value', False) == 'true' or \
                            change_set.get('management').get('new_value', False) == 'True' or \
                            change_set.get('management').get('new_value', False) == True:
                        n_management = True

                self.__apply_management_rules(name, c_management, n_management, change_set)

            if 'ipv6' in ch_set_keys:
                oper = change_set.get('ipv6').get('operation', None)
                n_ipv6 = False

                self.logger.debug('IPv6 is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    if change_set.get('ipv6').get('new_value', False) == 'true' or \
                            change_set.get('ipv6').get('new_value', False) == 'True' or \
                            change_set.get('ipv6').get('new_value', False) == True:
                        n_ipv6 = True
                    
                ppp_if.ipv6 = n_ipv6

            if 'ac-name' in ch_set_keys:
                oper = change_set.get('ac-name').get('operation', None )
                n_ac_name = None

                self.logger.debug('AC-name is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_ac_name = change_set.get('ac-name').get('new_value', None)

                ppp_if.ac_name = n_ac_name

    def __config_interface_wireguard(self, name, config, change_set):
        if self.get_network_mode() != 'vpp':
            self.logger.error('uCPE is not in VPP mode')
            return

        c_enabled = False
        c_l_port = None
        c_l_endpoint = None
        c_management = False
        c_local = None
        c_localv6 = None
        n_local = None
        n_localv6 = None

        self.logger.debug("Config {}".format(config))
        if config:
            
            c_enabled = config.get('enable', False)
            c_management = config.get('management', False)
            c_l_port = config.get('local-port', None)
            ipv4_pr = config.get('ipv4-prefix', None)
            ipv6_pr = config.get('ipv6-prefix', None)

            ifname = self.__get_ifname(name, None)

            if config.get('local-endpoint', None):
                c_l_endpoint = config.get('local-endpoint', None)
                c_local, c_localv6 = self._get_local_wireguard_endpoint_ip(c_l_endpoint)

            concentrator_endpoint = self._get_wireguard_concentrator_param(config, 'concentrator-endpoint')
            concentrator_password = self._get_wireguard_concentrator_param(config, 'concentrator-password')
            concentrator_port = self._get_wireguard_concentrator_param(config, 'concentrator-port')

            self.logger.error(f"PESHO IVO {concentrator_endpoint} {concentrator_password} {concentrator_port}")

        if not change_set and config:
            # Startup config
            self.logger.debug("APPLYING STARTUP FOR WIREGUARD INTERFACE - {}".format(name))
            wg_if = next(( a for a in self.wg_interfaces if a.name == config.get('name')), None)
            if wg_if:
                self.logger.debug("Interface \'{}\' allready exists.".format(wg_if.name))
            else:
                self.logger.debug("Add WireGuard configuration for \'{}\' interface" .format(config.get('name')))
                wg_if = wireguardif.WireGuardVPNInterface(config.get('name'), logger=self.logger, l_port=c_l_port,
                                                          l_endpoint=c_local, l_endpoint_v6=c_localv6, conc_endpoint=concentrator_endpoint,
                                                          conc_password=concentrator_password, conc_port=concentrator_port, enable=c_enabled)

                if 'peers' in config:
                    for peer in config['peers']:
                        self.logger.debug("Peer {}".format(peer))
                        r_endpoint = peer.get('peer-endpoint', None)
                        if not r_endpoint:
                            continue
                        r_port = peer.get('peer-port', None)
                        r_ips = peer.get('allowed-ips', None)
                        r_keepalive = peer.get('persistent-keepalive', 0)
                        wg_if.init_peer(r_endpoint=r_endpoint, r_port=r_port, allowed_ips=r_ips, persistent_keepalive=r_keepalive)

                self.wg_interfaces.append(wg_if)

                mtu = config.get('mtu', 1420)
                if mtu:
                    wg_if.mtu = int(mtu)

                if ipv4_pr:
                    wg_if.ip4_addr = ipaddress.IPv4Interface(str(ipv4_pr))

                if ipv6_pr:
                    wg_if.ip6_addr = ipaddress.IPv6Interface(str(ipv6_pr))

                # NOTE: ENABLE SETTER WAS USED HERE ... VERIFY IT WORKS !!!!!!!!!!!!!!!!!!!!!!!
                #wg_if.enable = c_enabled

                self.__add_del_ip_pass_tap_intf_if_wg_enabled(ifname, 1)
                self.__apply_management_rules(ifname, False, c_management, change_set)

        if change_set:
            self.logger.debug("APPLYING RUNNING FOR WIREGUARD INTERFACE - {}".format(name))
            # self.logger.debug('change_set {} ' .format(change_set))
            ch_set_keys = change_set.keys()
            self.logger.debug('Keys : {}'.format(ch_set_keys))

            wg_if = next(( a for a in self.wg_interfaces if a.name == name), None)

            if 'name' in ch_set_keys:
                if change_set.get('name', None):
                    if change_set.get('name').get('operation', None) == 'CREATED':
                        if wg_if:
                            self.logger.error("There is WireGuard interface object for {} ?!?!".format(name))
                        else:
                            wg_if = wireguardif.WireGuardVPNInterface( name, logger=self.logger )
                            self.wg_interfaces.append(wg_if)
                    elif change_set.get('name').get('operation', None) == 'DELETED':
                        if wg_if:
                            self.wg_interfaces.remove(wg_if)
                            del wg_if
                        return

            if 'local-port' in ch_set_keys:
                oper = change_set.get('local-port').get('operation', None )
                n_l_port = None

                self.logger.debug('Local port is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_l_port = change_set.get('local-port').get('new_value', None)

                wg_if.l_port = int(n_l_port)

            if 'local-endpoint' in ch_set_keys:
                oper = change_set.get('local-endpoint').get('operation', None )
                n_l_endpoint = None

                self.logger.debug('Local endpoint is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_l_endpoint = change_set.get('local-endpoint').get('new_value', None)

                n_local, n_localv6 = self._get_local_wireguard_endpoint_ip(n_l_endpoint)

                if n_local:
                    wg_if.l_endpoint = n_local
                    self.logger.debug('Local endpoint configured {}' .format(n_local))
                if n_localv6:
                    wg_if.l_endpoint_v6 = n_localv6
                    self.logger.debug('Local endpoint configured {}' .format(n_local))

            if 'mtu' in ch_set_keys:
                oper = change_set.get('mtu').get('operation', None )
                n_mtu = 1420

                self.logger.debug('MTU is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_mtu = change_set.get('mtu').get('new_value', 1420)

                wg_if.mtu = int(n_mtu)

            if 'enable' in ch_set_keys:
                oper = change_set.get('enable').get('operation', None)
                n_enable = False

                self.logger.debug('State is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    if change_set.get('enable').get('new_value', False) == 'true' or \
                            change_set.get('enable').get('new_value', False) == 'True' or \
                            change_set.get('enable').get('new_value', False) == True:
                        n_enable = True
                        self.__add_del_ip_pass_tap_intf_if_wg_enabled(ifname, 1)
                    else:
                        self.__add_del_ip_pass_tap_intf_if_wg_enabled(ifname, 0)
                    
                wg_if.enable = n_enable
            
            if 'ipv4-prefix' in ch_set_keys:
                oper = change_set.get('ipv4-prefix').get('operation', None)
                n_ip4_addr = None

                self.logger.debug('IPv4 prefix is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_ip4_addr = change_set.get('ipv4-prefix').get('new_value', None)
                    if n_ip4_addr:
                        n_ip4_addr = ipaddress.IPv4Interface(str(n_ip4_addr))
                    
                wg_if.ip4_addr = n_ip4_addr

            if 'ipv6-prefix' in ch_set_keys:
                oper = change_set.get('ipv6-prefix').get('operation', None)
                n_ip6_addr = None

                self.logger.debug('IPv4 prefix is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_ip6_addr = change_set.get('ipv6-prefix').get('new_value', None)
                    if n_ip6_addr:
                        n_ip6_addr = ipaddress.IPv6Interface(str(n_ip6_addr))
                    
                wg_if.ip6_addr = n_ip6_addr

            if 'management' in ch_set_keys:
                oper = change_set.get('management').get('operation', None)
                n_management = False

                self.logger.debug('Management is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    if change_set.get('management').get('new_value', False) == 'true' or \
                            change_set.get('management').get('new_value', False) == 'True' or \
                            change_set.get('management').get('new_value', False) == True:
                        n_management = True

                self.__apply_management_rules(name, c_management, n_management, change_set)

            if 'wireguard-concentrator/concentrator-endpoint' in ch_set_keys:
                oper = change_set.get('wireguard-concentrator/concentrator-endpoint').get('operation', None )
                c_endpoint = None

                self.logger.error('Concentrator endpoint is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    c_endpoint = change_set.get('wireguard-concentrator/concentrator-endpoint').get('new_value', None)

                c_local, c_localv6 = self._get_local_wireguard_endpoint_ip(c_endpoint)

                if c_local:
                    self.logger.error('Concentrator endpoint {}'.format(c_local))
                    wg_if.conc_endpoint = c_local
                    self.logger.error('Concentrator endpoint configured {}'.format(c_local))

            if 'wireguard-concentrator/concentrator-port' in ch_set_keys:
                oper = change_set.get('wireguard-concentrator/concentrator-port').get('operation', None )
                c_port = None

                self.logger.error('Concentrator port is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    c_port = change_set.get('wireguard-concentrator/concentrator-port').get('new_value', None)

                wg_if.conc_port = int(c_port)

            if 'wireguard-concentrator/concentrator-password' in ch_set_keys:
                oper = change_set.get('wireguard-concentrator/concentrator-password').get('operation', None )
                c_pass = None

                self.logger.error('Concentrator port is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    c_pass = change_set.get('wireguard-concentrator/concentrator-password').get('new_value', None)

                wg_if.conc_password = c_pass

            if str(change_set).find('peers') > -1:
                for xpath, change in change_set.items():
                    self.logger.debug('Peer xpath {} change {}' .format(xpath, change))
                    if "peers" not in xpath:
                        continue
                    # Get peer endpoint key
                    peer_key = change['xpath'][change['xpath'].find('peers[peer-endpoint=\'')+len('peers[peer-endpoint=\''):change['xpath'].rfind('\'')]
                    self.logger.debug('Peer config is changed for peer {}' .format(peer_key))
                
                    if xpath.find('/peer-endpoint') > -1:
                        # Peer created or deleted
                        oper = change['operation']
                        self.logger.debug('Peer peer-endpoint config is changed - operation {}' .format(oper))
                        if oper == 'MODIFIED' or oper == 'CREATED':
                            wg_if.add_peer_param(peer_key)
                        elif oper == 'DELETED':
                            wg_if.del_peer_param(peer_key)

                    if xpath.find('/peer-port') > -1:
                        oper = change['operation']
                        n_r_port = None
                        self.logger.debug('Peer peer-port config is changed - operation {}' .format(oper))
                        if oper == 'MODIFIED' or oper == 'CREATED':
                            n_r_port = change.get('new_value', None)
                        
                        wg_if.add_peer_param(peer_key, "r_port", n_r_port)

                    if xpath.find('/allowed-ips') > -1:
                        oper = change['operation']
                        n_r_nets = None
                        self.logger.debug('Peer allowed-ips config is changed - operation {}' .format(oper))
                        if oper == 'MODIFIED' or oper == 'CREATED':
                            n_r_nets = change.get('new_value', None)
                        
                        wg_if.add_peer_param(peer_key, "allowed_ips", n_r_nets)

                    if xpath.find('/persistent-keepalive') > -1:
                        oper = change['operation']
                        n_r_keepalive = None
                        self.logger.debug('Peer persistent-keepalive config is changed - operation {}' .format(oper))
                        if oper == 'MODIFIED' or oper == 'CREATED':
                            n_r_keepalive = change.get('new_value', None)
                        
                        wg_if.add_peer_param(peer_key, "persistent_keepalive", n_r_keepalive)

    def __interface_to_vsw_vpp(self, ifname, ifmode, vsw, apply_def=False):
        """ ifname: dataXX.vlan if vlan is configured
            vsw:    br-dataXX.vlan if vlan is configured
        """
        # if apply_def == False and 'br-'+ifname == vsw:
        #     return

        vpp_if = vppif.VppApiInterface('vsw-startup')
        vpp_iface = "host-" + ifname if ifmode == 'virtio' else ifname
        ifidx = vpp_if.get_ifidx_by_name(vpp_iface)
        bridx = vpp_if.get_bd_id_by_tag(vsw)
        if not bridx:
            if apply_def:
                ret = vpp_if.create_bridge_domain(ifidx, tag=vsw)
                bridx = vpp_if.get_bd_id_by_tag(vsw)
            else:
                del(vpp_if)
                return 1
        if ifidx and bridx:
            vpp_if.add_to_bridge(bridx,ifidx)
            self.logger.debug("Divert interface {} into BD with tag {}" .format(ifname, vsw))
        del(vpp_if)
        return 0

    def __vswitch_if_config(self, if_name, config, change_set, all_change_set):
        interface_type = config.get('type', None)
        if interface_type != 'ethernet' or if_name == 'mgmt0':
            return

        old_if_vlan = None
        if_def_vsw = 'br-' + if_name
        if_vsw = config.get('vswitch', None)
        if_vlan = config.get('vlan', 0)
        if_mode = config.get('mode', 'virtio')
        
        if change_set:
            self.logger.debug("APPLYING RUNNING VSWITCH FOR INTERFACE - {}".format(if_name))
            ch_set_keys = change_set.keys()

            if 'vswitch' in ch_set_keys:
                oper = change_set.get('vswitch').get('operation', None )
                
                self.logger.debug('VSwitch is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    if_vsw = change_set.get('vswitch').get('new_value', None)
                    
            if 'vlan' in ch_set_keys:
                oper = change_set.get('vlan').get('operation', None )
                
                self.logger.debug('Vlan is changed - operation {}' .format(oper))
                if oper == 'CREATED':
                    if_vlan = change_set.get('vlan').get('new_value', 0)
                    old_if_vlan = 0
                elif oper == 'MODIFIED':
                    if_vlan = change_set.get('vlan').get('new_value', 0)
                    old_if_vlan = change_set.get('vlan').get('old_value', 0)
                elif oper == 'DELETED':
                    if_vlan = 0
                    old_if_vlan = change_set.get('vlan').get('old_value', 0)

        if not if_vsw:
            self.logger.error("Missing VSwitch configuration")
            return
        old_vsw_vlan = None
        vsw_name = if_vsw.strip("br-")
        vsw_vlan = self.get_interface_param(vsw_name, 'vlan', 0)

        if all_change_set:
            ch_set_keys = None
            if vsw_name in all_change_set.keys():
                ch_set_keys = all_change_set.get(vsw_name, None)
            if ch_set_keys:
                if 'vlan' in ch_set_keys:
                    oper = ch_set_keys.get('vlan').get('operation', None )
                    self.logger.debug('Vlan for VSwitch interface {} is changed - operation {}' .format(vsw_name, oper))
                    if oper == 'CREATED':
                        vsw_vlan = ch_set_keys.get('vlan').get('new_value', 0)
                        old_vsw_vlan = 0
                    elif oper == 'MODIFIED':
                        vsw_vlan = ch_set_keys.get('vlan').get('new_value', 0)
                        old_vsw_vlan = ch_set_keys.get('vlan').get('old_value', 0)
                    elif oper == 'DELETED':
                        vsw_vlan = 0
                        old_vsw_vlan = ch_set_keys.get('vlan').get('old_value', 0)

        if self.get_network_mode() == 'ovs':
            self.logger.debug('Not implemented')
            return
        else:
            if self.vpp_not_running():
                self.logger.debug('VPP is not running')
                return
            if self.__interface_to_vsw_vpp(self.tag_interface(if_name, if_vlan), if_mode, self.tag_interface(if_vsw, vsw_vlan)) == 1:
                # new BD is still not created - move interface to the old one
                self.__interface_to_vsw_vpp(self.tag_interface(if_name, if_vlan), if_mode, self.tag_interface(if_vsw, old_vsw_vlan))
            if old_if_vlan != None:
                self.__interface_to_vsw_vpp(self.tag_interface(if_name, old_if_vlan), if_mode, self.tag_interface(if_def_vsw, old_if_vlan), apply_def=True)

    def __config_interface(self, name, config, change_set):
        enable = ""
        interface_type = self.get_interface_param(name, 'type', None)
        apn_changed = False
        user_changed = False
        ifmode_changed = False

        if not interface_type:
            if change_set:
                type_elem = change_set.get('type', None)
                if type_elem and type_elem.get('operation', None) == 'CREATED':
                    interface_type = type_elem.get('new_value', None)

        if not interface_type:
            self.logger.error('{} no type'.format(name))

        if interface_type == 'openvpn' or 'tun0' in name:
            self.__config_openvpn_interface(name, config, change_set)
            return

        if interface_type == 'ipsecvpn' or 'vti' in name:
            if change_set:
                self.__config_ipsecvpn_apply(name, change_set, self.ipsvpn_interfaces)
            else:
                self.__config_ipsecvpn_startup(name, config, self.ipsvpn_interfaces)
            return

        if interface_type == 'vxlan' and self.get_network_mode() == 'vpp':
            self.__config_interface_vxlan(name, config, change_set)
            return

        if interface_type == 'pppoe':
            self.__config_interface_pppoe(name, config, change_set)
            return

        if interface_type == 'loopback':
            self.__config_interface_loopback(name, config, change_set)
            return

        if interface_type == 'wireguard':
            self.__config_interface_wireguard(name, config, change_set)
            return

        if_names = [f for f in os.listdir('/sys/class/net') if
                    os.path.isfile(os.path.join('/sys/class/net', f)) or
                    os.path.islink(os.path.join('/sys/class/net', f))]

        basename = re.sub(r'(.*)(\.[0-9]+)$', r'\1', name)  # vlan tag interface - check for base name
        if basename not in if_names \
           and 'dpdk' != self.get_interface_param(name, 'mode', 'virtio') \
               and 'lte' != interface_type \
                   and 'wifi' != interface_type:
            self.logger.info("Interface {} not found in the system! Skipping configuration...".format(name))
            return

        if 'nat' == interface_type or 'nat' in name:
            self.__config_nat_interface(name, config, change_set)
            return

        if 'mgmt' in name:
            self.__config_mgmt_interface(name, config, change_set)
            return

        if not config:
            # Interface is deleted from config
            self.execute_shell_command("/scripts/apply_networking.py -z {}".format(name))
            return

        bootproto = self.get_interface_param(name, 'bootproto', 'dhcp')
        ipv4_prefix = self.get_interface_param(name, 'ipv4-prefix', '')
        ipv6_prefix = self.get_interface_param(name, 'ipv6-prefix', '')
        dns = self.get_interface_param(name, 'dns', '')
        dns6 = self.get_interface_param(name, 'dns6', '')
        gateway = self.get_interface_param(name, 'gateway', '')
        gateway6 = self.get_interface_param(name, 'gateway6', '')
        metric = self.get_interface_param(name, 'metric', 100)
        vlan = self.get_interface_param(name, 'vlan', 0)
        management = self.get_interface_param(name, 'management', '')
        management_old = 'false'

        ifname_old = self.__get_ifname(name, vlan)
        if 'inband' in name:
            ifname_old = name

        if change_set:
            for xpath, change in change_set.items():
                if xpath.find('vlan') > -1:
                    if change['operation'] != 'CREATED':
                        ifname_old = self.__get_ifname(name, change['old_value'])
                        if 'inband' in name:
                            ifname_old = name
                if change['operation'] == 'DELETED':
                    # Fill defaults
                    if xpath.find('bootproto') > -1:
                        bootproto = 'dhcp'
                    if xpath.find('ipv4-prefix') > -1:
                        ipv4_prefix = ''
                    if xpath.find('ipv6-prefix') > -1:
                        ipv6_prefix = ''
                    if xpath.find('dns6') > -1:
                        dns6 = ''
                    elif xpath.find('dns') > -1:
                        dns = ''
                    if xpath.find('metric') > -1:
                        metric = 100
                    if xpath.find('vlan') > -1:
                        self.handle_vpp_vlan_interface(name, change, change['operation'])
                        vlan = 0
                    if xpath.find('gateway6') > -1:
                        gateway6 = ''
                    elif xpath.find('gateway') > -1:
                        gateway = ''
                    if xpath.find('management') > -1:
                        management = 'false'
                    if xpath.find('apn') > -1:
                        config['apn'] = ""
                        apn_changed = True
                    if xpath.find('user') > -1:
                        user_changed = True
                        config['user'] = ""
                    if xpath.find('pass') > -1:
                        user_changed = True
                        config['pass'] = ""
                    continue
                if xpath.find('enable') > -1 and "tracehost" not in xpath:
                    enable = change['new_value']
                if xpath.find('bootproto') > -1:
                    bootproto = change['new_value']
                if xpath.find('ipv4-prefix') > -1:
                    ipv4_prefix = change['new_value']
                if xpath.find('ipv6-prefix') > -1:
                    ipv6_prefix = change['new_value']
                if xpath.find('dns6') > -1:
                    dns6 = change['new_value']
                elif xpath.find('dns') > -1:
                    dns = change['new_value']
                if xpath.find('metric') > -1:
                    metric = change['new_value']
                if xpath.find('vlan') > -1:
                    self.handle_vpp_vlan_interface(name, change, change['operation'])
                    vlan = change['new_value']
                if xpath.find('gateway6') > -1:
                    gateway6 = change['new_value']
                elif xpath.find('gateway') > -1:
                    gateway = change['new_value']
                if xpath.find('management') > -1:
                    management = change['new_value']
                    if 'old_value' in change:
                        management_old = change['old_value']
                if xpath.find('apn') > -1:
                    apn_changed = True
                    config['apn'] = change['new_value']
                if xpath.find('user') > -1:
                    user_changed = True
                    config['user'] = change['new_value']
                if xpath.find('pass') > -1:
                    user_changed = True
                    config['pass'] = change['new_value']
                # if "bootproto" in change["xpath"] and bootproto != "none" or "ipv4-prefix" in change["xpath"]:
                #         if os.path.exists('/config/central-box') or os.path.exists('/config/k8s-single-machine'):
                #             self.execute_shell_command("systemctl restart nfvtime-kubernetes-init.service")

        if enable == "false" and interface_type == "lte":
            self.logger.info("Disabling LTE iface")
            self._shell_return_ok("ifdown --ignore-error --force {}".format(ifname_old))
            self.execute_shell_command("/scripts/apply_networking.py -z {}".format(ifname_old))
            self._shell_return_ok("rm -rf /etc/network/interfaces.d/ifcfg-{}".format(ifname_old))
            return

        if enable == "false" and interface_type == "wifi":
            self.logger.info("Disabling wifi iface")
            self.execute_shell_command("systemctl stop wpa_supplicant.service hostapd.service")
            self.execute_shell_command("systemctl disable wpa_supplicant.service hostapd.service")
            self._shell_return_ok("ifdown --ignore-error --force {}".format(ifname_old))
            self.execute_shell_command("/scripts/apply_networking.py -z {}".format(ifname_old))
            self._shell_return_ok("rm -rf /etc/network/interfaces.d/ifcfg-{}".format(ifname_old))
            return

        if not ifmode_changed:
            ifmode = self.get_interface_param(name, 'mode', 'virtio')
        if ifmode == 'virtio' or ifmode == 'sr-iov':
            self.logger.info("Interface \'{}\' is in \'{}\' mode. Create minimal ifcfg file".format(name, ifmode))
            # options = ["-n {}".format(name), "-o {}".format('yes'), "-b manual"]
            options = ["-n {}".format(name), "-o {}".format('yes'), "-b manual", "-p \'sysctl -w net.ipv6.conf.{}.disable_ipv6=1\'".format(name)]
            if bootproto.find("6") < 0 or not ipv6_prefix:
                self._shell_return_ok("echo 0 > /proc/sys/net/ipv6/conf/{}/autoconf".format(name))
            self.execute_shell_command("/scripts/apply_networking.py {}".format(" ".join(options)))
            self._shell_return_ok("ifup --ignore-error {}".format(name))
            self.__set_igb_param(name, ifmode)

        if 'lte' == interface_type:
            self.logger.info("Interface \'{}\' is in LTE. Create minimal ifcfg file".format(name))
            options = ["-n {}".format(name), "-o {}".format('yes'), "-b manual"]
            self.execute_shell_command("/scripts/apply_networking.py {}".format(" ".join(options)))
            self._shell_return_ok("ifup --ignore-error {}".format(name))

        if 'wifi' == interface_type:
            self.logger.info("Interface {} is in wifi. Create minimal ifcfg file".format(name))
            # options = ["-n {}".format(name), "-o {}".format('yes'), "-b manual"]
            options = [ \
                        "-n {}".format(name), \
                        "-o {}".format('yes'), \
                        "-b manual", \
                        "-p \'sysctl --ignore -w net.ipv6.conf.{}.disable_ipv6=0\'".format(name), \
                        "-p \'sysctl --ignore -w net.ipv6.conf.br-{}.disable_ipv6=0\'".format(name), \
                       ]
            self.execute_shell_command("/scripts/apply_networking.py {}".format(" ".join(options)))
            self._shell_return_ok("ifup --ignore-error {}".format(name))
            wifi.config_wifi_interface(self, name, config, change_set)

        ifname = self.__get_ifname(name, vlan)
        if 'inband' in name:
            ifname = name

        if bootproto not in ['dhcp', 'dhcp6', 'dhcp46', 'manual', 'static']:
            bootproto = None

        if not bootproto:
            if ipv4_prefix or ipv6_prefix:
                bootproto = "static"
            else:
                bootproto = "manual"

        onbt = "yes"
        if 'lte' == interface_type:
            if not os.path.exists("/sys/class/net/{}".format(name)) or (enable != "true" and config['enable'] == False):
                onbt = "no"
            else:
                onbt = "yes"
        elif 'wifi' == interface_type:
            if not os.path.exists("/sys/class/net/{}".format(name)) or (enable != "true" and config['enable'] == False):
                onbt = "no"
            else:
                onbt = "yes"
        elif 'ethernet' == interface_type:
            onbt = "yes"

        self.logger.info('{} onbt is {}'.format(name, onbt))

        self.__config_interface_ippass(name, config, change_set)

        # Reconnect all ipsecvpn that have local IP equal to IP of adaptor that is going to be down. Disconnect first.
        ipsecvpn_reenable_list = self.__ipsecvpn_for_reenable(ifname_old)
        ipsecvpn_reenable_count = len(ipsecvpn_reenable_list)
        if ipsecvpn_reenable_count > 0:
            self.logger.debug("{} ipsecvpn(s) need to be reenabled!".format(ipsecvpn_reenable_count))
            for ipsec_iface in ipsecvpn_reenable_list:
                self.__ipsecvpn_disconnect(ipsec_iface)

        self._shell_return_ok("ifdown --ignore-error --force {}".format(ifname_old))
        self._shell_return_ok("rm -rf /etc/network/interfaces.d/ifcfg-{}".format(ifname_old))

        if 'lte' == interface_type:
            from tveif import if_lte
            if apn_changed or user_changed:
                self._shell_return_ok("rm -f /var/lib/dhcp/dhclient.{}.leases".format(ifname_old))
            modem_type=if_lte.lte_get_driver_type(config["name"])
            if modem_type == 1:
                if_lte.reconnect_lte_qmi_dev(config, if_lte.modem_default)
            elif modem_type == 2 or modem_type == 4:
                if_lte.setup_lte_mbim_dev(config,if_lte.mbim_modem_default) 
                

        self.__apply_management_rules(ifname, management_old, management, change_set)

        options = ["-n {}".format(ifname), "-o {}".format(onbt)]
        if bootproto:
            options.append("-b {}".format(bootproto))
        if ipv4_prefix:
            options.append("--ipv4-prefix={}".format(ipv4_prefix))
        if ipv6_prefix:
            options.append("--ipv6-prefix={}".format(ipv6_prefix))
            options.append('-p"echo 0 > /proc/sys/net/ipv6/conf/%s/accept_dad"' % str(ifname))
        if dns:
            options.append("-d {}".format(dns))
        if dns6:
            options.append("--ipv6-dns={}".format(dns6))
        if gateway:
            options.append("-g {}".format(gateway))
        if gateway6:
            options.append("--ipv6-gateway={}".format(gateway6))
            options.append('-j"ip -6 route add {gateway} dev {interface} && ip -6 route add default via {gateway}"'.format(gateway=gateway6, interface=ifname))
            options.append('-k"ip -6 route del {gateway} dev {interface} && ip -6 route del default via {gateway}"'.format(gateway=gateway6, interface=ifname))
        if metric:
            options.append("-c {}".format(metric))
        if ifname.find("br") > -1:
            options.append("--ovs-master ovs --ovs-ports {}".format(name))
            if bootproto.find("6") < 0 or not ipv6_prefix:
                self._shell_return_ok("echo 0 > /proc/sys/net/ipv6/conf/{}/autoconf".format(ifname))

        if not (ifmode == 'dpdk' and self.hw_type == 'cybertan8040'):
            self.execute_shell_command("/scripts/apply_networking.py {}".format(" ".join(options)))

        try:
            if 'data' in ifname and not self.vpp_not_running():
                vppintname=name
                if ifmode=='virtio':
                    cmd='A=$(cat /sys/class/net/%s/address); B=$(cat /sys/class/net/%s/address) ; if [ -f /sys/class/net/%s/address ] && [ -n "$B" ] && [ -n "$A" ] && [ "$A" != "$B" ] ; then ifconfig %s hw ether $A; logger "Restoring MAC of %s";  fi' % (name, ifname, ifname, ifname, ifname )
                else:
                    cmd='A=$(vppctl show hard %s | grep Ethern | xargs | cut -f3 -d" "); B=$(cat /sys/class/net/%s/address) ; if [ -f /sys/class/net/%s/address ] && [ -n "$B" ] && [ -n "$A" ] && [ "$A" != "$B" ] ; then ifconfig %s hw ether $A; logger "Restoring MAC of %s";  fi' % (vppintname, ifname, ifname, ifname, ifname )
                os.system(cmd)
        except:
            pass

        if onbt == "yes":
            # On HawkEye 6010 DPDK interface are visible to kernel
            # and bringing up bridge interface does not bring up
            # physical interface. So do it here instead.
            if (self.hw_type in ["hk6010", "hk6010-8g", "hk6010-16g"]
                    and ifmode == 'dpdk'):
                attempts=0
                while attempts < 5:
                    attempts  += 1
                    try:
                        subprocess.check_call(
                            ['ip', 'link', 'set', name, 'up'])
                        break
                    except:
                        sleep(2)
            try:
                if not (ifmode == 'dpdk' and self.hw_type == 'cybertan8040'):
                    self.logger.info("ifup interface {}".format(ifname))
                    subprocess.check_call(['ifup', '--ignore-error', ifname])
            except subprocess.CalledProcessError as ex:
                self.logger.error('ifup command failed!')
                #even if up command fail, continue wuth the other interfaces.
                #self._shell_return_ok("ifdown --ignore-error --force {}".format(ifname))
                #raise IfUpError('ifup error: {}'.format(ex), 1)

        if dns:
            dnses = dns.split(',')
            dnsopts = ''.join(map(lambda dns_srv: ' '+ dns_srv, dnses))            
            subprocess.call('/usr/bin/resolvectl -4 dns {} {}'.format(name, dnsopts), shell=True)
        if dns6:
            dns6es = dns6.split(',')
            dns6opts = ''.join(map(lambda dns_srv: ' '+ dns_srv, dns6es))            
            subprocess.call('/usr/bin/resolvectl -6 dns {} {}'.format(ifname, dns6opts), shell=True)

        if ipv4_prefix or ipv6_prefix:
            self.handle_vpp_interface_addresses(ifname, ipv4_prefix, ipv6_prefix)
            if self.get_kubernetes_bind_interface() == name:
                # We prioritize IPv6
                ip_prefix = ipv6_prefix or ipv4_prefix
                ip = ip_prefix.split('/')[0]
                #some legacy code left here, lets notify it also
                subprocess.Popen(['/scripts/set_openstack_address.sh', f'{ip}'])
                subprocess.Popen(['/scripts/handle_k8s_ip_change.sh', f'{ip}'])
                #self.execute_shell_command(f"/scripts/handle_k8s_ip_change.sh {ipv4} &")
                

        # Reconnect all ipsecvpn that have local IP equal to IP of adaptor that is going to be down. Connect.
        for ipsec_iface in ipsecvpn_reenable_list:
            self.__ipsecvpn_connect(ipsec_iface)

        ipsecvpn_reenable_list = []

    def __apply_management_rules(self, ifname, management_old, management, change_set):
        if not ifname:
            self.logger.debug('No ifname')
            return

        if management_old == 'true' or management_old == "True" or change_set == None:
            if management == 'false' or management == "False":
                ssh_port = self._get_ssh_port()
                self._shell_return_ok("iptables -I INPUT 1 -i {} -p tcp -m tcp --dport {} -j REJECT --reject-with tcp-reset".format(ifname, ssh_port))
                self._shell_return_ok("ip6tables -I INPUT 1 -i {} -p tcp -m tcp --dport {} -j REJECT --reject-with tcp-reset".format(ifname, ssh_port))
                self._shell_return_ok("iptables -I INPUT 1 -i {} -p tcp -m tcp --dport 80 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("ip6tables -I INPUT 1 -i {} -p tcp -m tcp --dport 80 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("iptables -I INPUT 1 -i {} -p tcp -m tcp --dport 443 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("ip6tables -I INPUT 1 -i {} -p tcp -m tcp --dport 443 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("iptables -I INPUT 1 -i {} -p tcp -m tcp --dport 830 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("ip6tables -I INPUT 1 -i {} -p tcp -m tcp --dport 830 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("iptables -I INPUT 1 -i {} -p tcp -m tcp --dport 5000 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("ip6tables -I INPUT 1 -i {} -p tcp -m tcp --dport 5000 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("iptables -I INPUT 1 -i {} -p tcp -m tcp --dport 7373 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("ip6tables -I INPUT 1 -i {} -p tcp -m tcp --dport 7373 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("iptables -I INPUT 1 -i {} -p tcp -m tcp --dport 35357 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("ip6tables -I INPUT 1 -i {} -p tcp -m tcp --dport 35357 -j REJECT --reject-with tcp-reset".format(ifname))

        if management_old == 'false' or management_old == "False":
            if management == 'true' or management == "True":
                ssh_port = self._get_ssh_port()
                self._shell_return_ok("iptables -D INPUT -i {} -p tcp -m tcp --dport {} -j REJECT --reject-with tcp-reset".format(ifname, ssh_port))
                self._shell_return_ok("ip6tables -D INPUT -i {} -p tcp -m tcp --dport {} -j REJECT --reject-with tcp-reset".format(ifname, ssh_port))
                self._shell_return_ok("iptables -D INPUT -i {} -p tcp -m tcp --dport 80 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("ip6tables -D INPUT -i {} -p tcp -m tcp --dport 80 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("iptables -D INPUT -i {} -p tcp -m tcp --dport 443 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("ip6tables -D INPUT -i {} -p tcp -m tcp --dport 443 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("iptables -D INPUT -i {} -p tcp -m tcp --dport 830 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("ip6tables -D INPUT -i {} -p tcp -m tcp --dport 830 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("iptables -D INPUT -i {} -p tcp -m tcp --dport 5000 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("ip6tables -D INPUT -i {} -p tcp -m tcp --dport 5000 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("iptables -D INPUT -i {} -p tcp -m tcp --dport 7373 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("ip6tables -D INPUT -i {} -p tcp -m tcp --dport 7373 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("iptables -D INPUT -i {} -p tcp -m tcp --dport 35357 -j REJECT --reject-with tcp-reset".format(ifname))
                self._shell_return_ok("ip6tables -D INPUT -i {} -p tcp -m tcp --dport 35357 -j REJECT --reject-with tcp-reset".format(ifname))

    def __start_openvpn(self):
        self.execute_shell_command("systemctl enable openvpn@client.service")
        if not os.path.isfile('/config/.hamode'):
            try:
                self.execute_shell_command("systemctl --no-block start openvpn@client.service")
            except:
                self.logger.warning("Can not start openvpn")

    def __stop_openvpn(self):
        try:
            self.execute_shell_command("systemctl disable openvpn@client.service")
            self.execute_shell_command("systemctl --no-block stop openvpn@client.service")
            self.execute_shell_command("rm -f /etc/network/interfaces.d/ifcfg-tun0")
        except:
            self.logger.warning("Can not stop openvpn")

    def __restart_openvpn(self):
        do_restart = True
        
        if os.path.isfile('/config/.hamode'):
            do_restart = False
            try:
                if str(self.execute_shell_command("systemctl is-active openvpn@client.service")).strip("\n") == "active":
                    do_restart = True
            except:
                pass
                
        if not do_restart:
            return

        try:
            self.execute_shell_command("systemctl stop openvpn@client.service")
            self.execute_shell_command("systemctl --no-block start openvpn@client.service")
        except:
            self.logger.warning("Can not restart openvpn")

    def _request_tve_reboot(self):
        # Creation of TVE_REBOOT_DIR directory triggers the restart.
        # If the dir is present then remove it first, then create it. Otherwise the trigger will not sense the creation
        if os.path.exists(TVE_REBOOT_DIR):
            os.rmdir(TVE_REBOOT_DIR)

        # Create dirs
        os.makedirs(TVE_REBOOT_DIR)

    def __post_interface_config(self):
        # try:
        #     pid = subprocess.check_output(["/bin/systemctl", "show", "trace_host", "-p", "MainPID", "--value"]).decode()
        #     if int(pid) != 0:
        #         subprocess.check_call(["/bin/kill", "-s", "SIGUSR1", pid.rstrip()])
        # except:
        #     self.logger.debug("systemctl start trace_host failed!")
        #     pass
        if self.igb_recalc_needed:
            self._recalc_igb_params()
            if self.initramfs_changed:
                self.logger.warning("Requesting system reboot.[initramfs change]")
                self._request_tve_reboot()

    def __pre_interface_config(self):
        try:
            pass
        except:
            self.logger.debug("systemctl stop trace_host failed!")
            pass

    def __str_get_ifname(self, str):
        r = re.compile(".*((data[0-9]{1,2})|(lte[0-9]{1,2})|(nat[0-9]{1,2})|(wifi0)|(mgmt[0-9]{1})).*")
        m = r.match(str)
        # self.logger.info('{} - {}'.format(str, None if m is None else m.groups()))
        if m is None:
            return None

        for name in m.groups():
            if name is not None:
                return name

        return None

    def __get_ifname_from_change(self, change_elem):
        name_id = change_elem['xpath'].lstrip(self.namespace).split('/', 1)
        # self.logger.info(name_id)
        if name_id:
            return self.__str_get_ifname(name_id[0])

        return None

    def __is_nat_change(self, change):
        name_element = change['xpath'].lstrip(self.namespace).split('/', 1)
        if name_element:
            r = re.compile(".*nat[0-9]{1,2}.*")
            m = r.match(name_element[0])
            if m is None:
                return False
            else:
                return True
        return False

    def __is_wifi_nat_change(self, change):
        name_element = change['xpath'].lstrip(self.namespace).split('/', 1)
        if name_element:
            r = re.compile(".*wifi[0-9]{1,2}.*")
            m = r.match(name_element[0])
            if m is None:
                return False
            else:
                return True
        return False

    def __apply_nat_changes(self, change_set, nats):
        nat_change_set = filter(lambda elem: self.__is_nat_change(elem), change_set)
        if not nat_change_set:
            return

        for nat_change in nat_change_set:
            if_nat = self.__get_ifname_from_change(nat_change)

            # is wan changed
            change_path = nat_change['xpath'].lstrip(self.namespace).split('/', 1)
            if len(change_path) > 1:
                if change_path[1].startswith('wan-interface'):
                    if nat_change['operation'] == 'CREATED':
                        if nat_change['new_value']:
                            nat_to_ch = [x for x in nats if x['name'] == if_nat]
                            if nat_to_ch:
                                nat_wan_list = nat_to_ch[0].get('wan-interface', [])
                                nat_wan_list.append({'name': nat_change['new_value']})
                                nat_to_ch[0]['wan-interface'] = nat_wan_list

                    if nat_change['operation'] == 'DELETED':
                        if nat_change['old_value']:
                            nat_to_ch = [x for x in nats if x['name'] == if_nat]
                            if nat_to_ch:
                                nat_wan_list = nat_to_ch[0].get('wan-interface', [])
                                if nat_wan_list:
                                    nat_wan_list.remove({'name': nat_change['old_value']})

    def __wan_handler(self, change_set):
        running = self.get_running_config(full_path=False)
        for iftype in ['nat', 'wifi']:
            if 'nat' in iftype:
                nats = filter(lambda elem: elem['type'] == 'nat', running)
                self.__apply_nat_changes(change_set, nats)
                # self.logger.info(nats)
                wans = filter(lambda elem: ((elem['type'] == 'ethernet' or elem['type'] == 'lte' or elem['type'] == 'wifi' or elem['type'] == 'pppoe') and elem['name'] != 'mgmt0'), running)
            if 'wifi' in iftype:
                nats = filter(lambda elem: elem['type'] == 'wifi', running)
                self.__apply_wifi_nat_changes(change_set, nats)
                # self.logger.info(nats)
                wans = filter(lambda elem: ((elem['type'] == 'ethernet' or elem['type'] == 'lte') and elem['name'] != 'mgmt0'), running)

            for change_elem in change_set:
                if_name = self.__get_ifname_from_change(change_elem)
                if if_name not in [x['name'] for x in wans]:
                    # This interface cannot be wan - skip
                    # self.logger.info('Skip {} - now WAN'.format(if_name))
                    continue

                # self.logger.info('Change for interface {}'.format(if_name))
                path = change_elem['xpath'].lstrip(self.namespace).split('/')
                path.pop(0)
                if len(path) > 0 and path[0] == 'vlan':
                    if 'nat' in iftype:
                        wan_in = filter(lambda elem: elem.get('wan-interface', None) is not None and if_name in [x['name'] for x in elem.get('wan-interface', None)],nats)
                    if 'wifi' in iftype:
                        wan_in = filter(lambda elem: elem.get('wifi-wan-interface', None) is not None and if_name in [x['name'] for x in elem.get('wifi-wan-interface', None)],nats)

                    # self.logger.info(wan_in)
                    for nat in wan_in:
                        old_tag = None
                        new_tag = None
                        if change_elem['operation'] == 'MODIFIED':
                            old_tag = int(change_elem['old_value'])
                            new_tag = int(change_elem['new_value'])

                        if change_elem['operation'] == 'DELETED':
                            old_tag = int(change_elem['old_value'])
                            new_tag = None

                        if change_elem['operation'] == 'CREATED':
                            old_tag = None
                            new_tag = int(change_elem['new_value'])

                        self._clean_nat_iptables(nat['name'], if_name, wan_tag=old_tag)
                        self.__clean_wan_routes(nat['name'], if_name, wan_tag=old_tag)
                        self.__set_nat_iptables(nat['name'], if_name, wan_tag=new_tag)
                        self.__set_wan_routes(nat['name'], if_name, wan_tag=new_tag)

    def __apply_wifi_nat_changes(self, change_set, nats):
        nat_change_set = filter(lambda elem: self.__is_wifi_nat_change(elem), change_set)
        if not nat_change_set:
            return

        for nat_change in nat_change_set:
            if_nat = self.__get_ifname_from_change(nat_change)

            # is wan changed
            change_path = nat_change['xpath'].lstrip(self.namespace).split('/', 1)
            if len(change_path) > 1:
                if change_path[1].startswith('wifi-wan-interface'):
                    if nat_change['operation'] == 'CREATED':
                        if nat_change['new_value']:
                            nat_to_ch = [x for x in nats if x['name'] == if_nat]
                            if nat_to_ch:
                                nat_wan_list = nat_to_ch[0].get('wifi-wan-interface', [])
                                nat_wan_list.append({'name': nat_change['new_value']})
                                nat_to_ch[0]['wifi-wan-interface'] = nat_wan_list

                    if nat_change['operation'] == 'DELETED':
                        if nat_change['old_value']:
                            nat_to_ch = [x for x in nats if x['name'] == if_nat]
                            if nat_to_ch:
                                nat_wan_list = nat_to_ch[0].get('wifi-wan-interface', [])
                                if nat_wan_list:
                                    nat_wan_list.remove({'name': nat_change['old_value']})

    def apply_startup_config(self, change_set=None):
        config_indexed = self.get_interfaces(startup=True)
        # nat_routing_flag = False
        for name, config in sorted(config_indexed.items()):
            # if config['type'] == "ipsecvpn":
            #     self.ipsvpn_interfaces.append(ips.ipsec_vpn_if(name))

            self.__config_interface(name, config, change_set)
            self.__vswitch_if_config(name, config, None, None)
            # TODO (sandev): Thread interfaces' configurations
            """
            self.logger.info(f"DBG: Starting thread for: {name}")
            if "br-" not in name:
                try:
                    thread = threading.Thread(target=self.__config_interface, args=(name, config, change_set,), daemon=True)
                    thread.start()
                except Exception as e:
                    self.logger.error("Error configuring interface: {name}")
                    self.logger.error(repr(e))
            else:
                config_bridge_interfaces[name] = config
            """

        self.__post_interface_config()
        radvd.init(interfaces=self.radvd_ifs, logger=self.logger)
        if self.tracehost is not None:
            if (self.tracehost.load_initial() > 0):
                self.logger.error("Failed to load initial tracehost interfaces configuration")

        return True

    def apply_config(self, change_set=None, apply_startup=False):
        if apply_startup:
            return self.apply_startup_config(change_set)

        # Get current running
        config_indexed = self.get_interfaces()

        self.__pre_interface_config()
        interface_changes = {}
        for change in change_set:
            if not change["xpath"].startswith(self.namespace):
                continue

            if change['operation']=='CREATED' and change['new_value']=="":
                if change["xpath"].endswith("/dns") or change["xpath"].endswith("/gateway") or change["xpath"].endswith("/trace-host") or change["xpath"].endswith("/apn"):
                        continue

            name = str(change['xpath'])
            name = name[name.find("interface[name='") + len("interface[name='"):]
            name = name[:name.find("]") - 1]
            if name not in interface_changes:
                interface_changes[name] = {}

            interface_changes[name][change['xpath'][(change['xpath'].find(']/') + len(']/')):]] = change

        try:
            down_interfaces = []
            # ifdown the interfaces which metrics have been modified
            for name, xpath_dict in interface_changes.items():
                if name in config_indexed and self.tracehost is not None and ('data' in name or 'lte' in name or 'wifi' in name or 'ppp' in name):
                    # if 'tracehost' in str(xpath_dict.keys()) or 'enable' in str(xpath_dict.keys()):
                    self.tracehost.del_config(name)
                if "metric" in xpath_dict:
                    vlan = self.get_interface_param(name, 'vlan', 0)
                    sys_name = self.__get_ifname(name, vlan)
                    self.execute_shell_command(f"ifdown  {sys_name} || true")
                    down_interfaces.append(sys_name)

            # Go over modified interfaces
            for name, xpath_dict in interface_changes.items():
                if name in config_indexed:
                    self.__config_interface(name, config_indexed[name], xpath_dict)
                    if 'vlan' in xpath_dict.keys() or 'vswitch' in xpath_dict.keys():
                        self.__vswitch_if_config(name, config_indexed[name], xpath_dict, interface_changes)
                else:
                    if 'vti' in name:
                        self.ipsvpn_interfaces.append(ips.ipsec_vpn_if(name))
                    if 'ppp' in name:
                        if 'ethernet-interface' in xpath_dict.keys():
                            oper = xpath_dict.get('ethernet-interface').get('operation', None)
                            if oper == 'MODIFIED' or oper == 'CREATED':
                                ethif = xpath_dict.get('ethernet-interface').get('new_value', None)
                                if ethif in interface_changes.keys():
                                    if 'vlan' in interface_changes.get(ethif, None).keys():
                                        xpath_dict.update({'eth-if-vlan': interface_changes.get(ethif).get('vlan')})
                    if 'vxlan' in name:
                        if 'local/vtep-if' in xpath_dict.keys():
                            oper = xpath_dict.get('local/vtep-if').get('operation', None)
                            if oper == 'MODIFIED' or oper == 'CREATED':
                                ethif = xpath_dict.get('local/vtep-if').get('new_value', None)
                                if ethif in interface_changes.keys():
                                    if 'ipv4-prefix' in interface_changes.get(ethif, None).keys():
                                        xpath_dict.update({'local/lvtep-if-ipv4-prefix': interface_changes.get(ethif)
                                                          .get('ipv4-prefix')})

                    self.__config_interface(name, None, xpath_dict)

            # ifup the interfaces which metrics have been changed
            if down_interfaces:
                self.execute_shell_command(f"ifup {' '.join(down_interfaces)}  || true")
                del down_interfaces

        except IfUpError as ex:
            self.logger.error(ex)
            return False

        radvd.init(interfaces=self.radvd_ifs, logger=self.logger)

        # All done. Save the config!
        self.__wan_handler(change_set)
        return True

    def post_apply_config(self, change_set):
        if not change_set:
            return

        trace_host_interfaces = []

        for change in change_set:
            name = str(change['xpath'])
            name = name[name.find("interface[name='") + len("interface[name='"):]
            name = name[:name.find("]") - 1]
            iftype = ""
            if name:
                iftype = self.get_interface_param(name, 'type')

            if iftype == 'ethernet' and change['xpath'].find('mode') > -1:
                self.logger.warning("Interface mode changed to {}. Requesting system reboot.".format(change['new_value']))
                self._request_tve_reboot()

            if (iftype == 'ethernet' or iftype == 'lte' or iftype == 'wifi' or iftype == 'pppoe') and change['xpath'].find('ip-passthrough') > -1:
                if (('old_value' in change and 'new_value' in change and (str(change['old_value']) != str(change['new_value']))) or
                    ('old_value' not in change and 'new_value' in change and (str(change['new_value']) == "true" or change['new_value'] == True) ) or
                    ('old_value' in change and 'new_value' not in change and (str(change['old_value']) == "true" or change['old_value'] == True) ) ):
                    self.logger.warning("Interface {} - IP Passthrough changed to {}. Requesting system reboot.".format(name, change['new_value']))
                    self._request_tve_reboot()
            if iftype == 'ethernet' or iftype == 'lte' or iftype == 'wifi' or iftype == 'pppoe':
                trace_host_interfaces.append(name)

        wifi.post_apply_config_wifi(self, change_set)
        self.__post_interface_config()
        if trace_host_interfaces and self.tracehost is not None:
            for ifname in trace_host_interfaces:
                if (self.tracehost.load_config(ifname) > 0):
                    self.logger.error("Failed to load interface {} config" .format(ifname))

    def status_get_cb(self, xpath, private_data):
        data = {'ucpe:status': {'interfaces:interface': []}}
        ipsvpn_fields = ['conname','state','established','rekey-time','vips','nat-local','nat-any','vips','local','remote','local-id',
                         'remote-id','version','integ-alg','encr-alg','encr-ksize','dh-group']
        ipsvpn_ch_fields = ['chname','state','bytes-in','bytes-out','pkt-in','pkt-out','mode','proto','local-ts','remote-ts','integ-alg',
                            'encr-alg','encr-ksize','dh-group']

        for vpn in self.ipsvpn_interfaces:
            sa = vpn.get_sa(all_child=True)
            iface_data = {'name': str(vpn.name)}

            if sa is not None:
                for field in ipsvpn_fields:
                    iface_data[str(field)] = str(sa.get(field, ""))

                chs = sa['child']
                if chs is not None:
                    iface_data['child'] = []
                    for ch in chs:
                        ch_data = {'chname': str(ch.get('chname'))}
                        for ch_field in ipsvpn_ch_fields:
                            ch_data[str(ch_field)] = str(ch.get(ch_field, ""))
                        iface_data['child'].append(ch_data)
            data['ucpe:status']['interfaces:interface'].append(iface_data)

        if len(data['ucpe:status']['interfaces:interface']) == 0:
            return None
        return data

    # Client helper functions

    def get_network_mode(self):
        return self.get_item("/ucpe:config/system:system/network-mode")

    def get_kubernetes_bind_interface(self):
        return self.get_item("/ucpe:config/kubernetes:kubernetes/bind-interface")

    def set_interface(self, name):
        self.set_item("{}[name='{}']".format(self.namespace, name), name)

    def get_interface(self, name):
        return self.get_item("{}[name='{}']".format(self.namespace, name))

    def get_interfaces(self, startup=False):
        config_indexed = {}
        sys_conf = self.get_startup_config(full_path=False) if startup else self.get_running_config(full_path=False)
        for interface in sys_conf:
            config_indexed[interface['name']] = interface
        return config_indexed

    def unset_interface(self, name):
        self.delete_item("{}[name='{}']".format(self.namespace, name))

    def set_interface_param(self, name, param, value):
        self.set_item("{}[name='{}']/{}".format(self.namespace, name, param), value)

    def get_interface_param(self, name, param, default=None):
        try:
            for interface in self.get_running_config(full_path=False):
                if interface['name'] == name and param in interface:
                    return self.get_item("{}[name='{}']/{}".format(self.namespace, name, param))
        except:
            return default

    def unset_interface_param(self, name, param):
        self.delete_item("{}[name='{}']/{}".format(self.namespace, name, param))

    def set_nat_wan_interface(self, nat_name, name):
        self.set_item("{}[name='{}']/wan-interface[name='{}']".format(self.namespace, nat_name, name), name)

    def get_nat_wan_interfaces(self, name):
        config_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'wan-interface' in interface:
                    for wan in interface['wan-interface']:
                        config_indexed[wan['name']] = wan
        return config_indexed

    def unset_nat_wan_interface(self, nat_name, name):
        self.delete_item("{}[name='{}']/wan-interface[name='{}']".format(self.namespace, nat_name, name))

    def unset_nat_wan_interfaces(self, nat_name):
        self.delete_item("{}[name='{}']/wan-interface".format(self.namespace, nat_name))

    def set_concentrator_endpoint(self, ifname, value):
        self.set_item("{}[name='{}']/ovpn-cfg/concentrator-endpoint".format(self.namespace, ifname), value)

    def get_concentrator_endpoint(self, ifname):
        return self.get_item("{}[name='{}']/ovpn-cfg/concentrator-endpoint".format(self.namespace, ifname))

    def set_concentrator_port(self, ifname, value):
        self.set_item("{}[name='{}']/ovpn-cfg/concentrator-port".format(self.namespace, ifname), value)

    def get_concentrator_port(self, ifname):
        return self.get_item("{}[name='{}']/ovpn-cfg/concentrator-port".format(self.namespace, ifname))

    def get_ippass_param(self, ifname):
        return self.get_item("{}[name='{}']/ip-passthrough".format(self.namespace, ifname))

    def set_ippass_param(self, ifname, value):
        return self.set_item("{}[name='{}']/ip-passthrough".format(self.namespace, ifname), value)

    # Define get, set and unset methods for IKE1 local round parameters
    def unset_ike1_round(self, ifname, type, id):
        self.delete_item("{}[name='{}']/ike-phase-1/{}/rounds[id='{}']".format(self.namespace, ifname, type, id))

    def set_ike1_round(self, ifname, type, id):
        self.set_item("{}[name='{}']/ike-phase-1/{}/rounds[id='{}']".format(self.namespace, ifname, type, id), id)

    def set_ike1_round_param(self, ifname, type, id, param, value):
        self.set_item("{}[name='{}']/ike-phase-1/{}/rounds[id='{}']/{}".format(self.namespace, ifname, type, id, param), value)

    def unset_ike1_round_param(self, ifname, type, id, param):
        self.delete_item("{}[name='{}']/ike-phase-1/{}/rounds[id='{}']/{}".format(self.namespace, ifname, type, id, param))

    def unset_ike1_local_round(self, ifname, id):
        self.delete_item("{}[name='{}']/ike-phase-1/local/rounds[id='{}']".format(self.namespace, ifname, id))

    def unset_ike1_local_rounds(self, ifname):
        self.delete_item("{}[name='{}']/ike-phase-1/local/rounds".format(self.namespace, ifname))

    def set_ike1_local_round(self, ifname, id):
        self.set_item("{}[name='{}']/ike-phase-1/local/rounds[id='{}']".format(self.namespace, ifname, id), id)

    def get_ike1_local_rounds(self, name):
        local_rounds_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'ike-phase-1' in interface:
                    for config in interface['ike-phase-1']:
                        if 'local' in config and 'rounds' in interface['ike-phase-1']['local']:
                            local_rounds_indexed = interface['ike-phase-1']['local']['rounds']
        return local_rounds_indexed

    def set_ike1_local_round_param(self, ifname, id, param, value):
        self.set_item("{}[name='{}']/ike-phase-1/local/rounds[id='{}']/{}".format(self.namespace, ifname, id, param), value)

    # Define get, set and unset methods for IKE1 remote round parameters
    def unset_ike1_remote_round(self, ifname, id):
        self.delete_item("{}[name='{}']/ike-phase-1/remote/rounds[id='{}']".format(self.namespace, ifname, id))

    def unset_ike1_remote_rounds(self, ifname):
        self.delete_item("{}[name='{}']/ike-phase-1/remote/rounds".format(self.namespace, ifname))

    def set_ike1_remote_round(self, ifname, id):
        self.set_item("{}[name='{}']/ike-phase-1/remote/rounds[id='{}']".format(self.namespace, ifname, id), id)

    def get_ike1_remote_rounds(self, name):
        remote_rounds_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'ike-phase-1' in interface:
                    for config in interface['ike-phase-1']:
                        if 'remote' in config and 'rounds' in interface['ike-phase-1']['remote']:
                            remote_rounds_indexed = interface['ike-phase-1']['remote']['rounds']
        return remote_rounds_indexed

    def set_ike1_remote_round_param(self, ifname, id, param, value):
        self.set_item("{}[name='{}']/ike-phase-1/remote/rounds[id='{}']/{}".format(self.namespace, ifname, id, param), value)

    # Define get, set and unset methods for IKE1 Diffie-Hellman groups
    def set_dh_groups(self, phase, ifname, group):
        try:
            self.set_item("{}[name='{}']/{}/dh-groups[group='{}']".format(self.namespace, ifname, phase, group), group)
        except Exception as ex:
            self.logger.error('Set Diffie-Hellman groups "{group}" failed. Not valid group. {msg}'.format(group=group, msg=str(ex)))
            print('Set Diffie-Hellman groups "{group}" failed. Not valid group. {msg}'.format(group=group, msg=str(ex)))

    def set_ike1_dh_groups(self, ifname, group):
        try:
            self.set_item("{}[name='{}']/ike-phase-1/dh-groups[group='{}']".format(self.namespace, ifname, group), group)
        except Exception as ex:
            self.logger.error('Set Diffie-Hellman groups "{group}" failed. Not valid group. {msg}'.format(group=group, msg=str(ex)))
            print('Set Diffie-Hellman groups "{group}" failed. Not valid group. {msg}'.format(group=group, msg=str(ex)))

    def get_ike1_dh_groups(self, name):
        groups_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'ike-phase-1' in interface:
                    for config in interface['ike-phase-1']:
                        if 'dh-groups' in config:
                            groups_indexed = interface['ike-phase-1']['dh-groups']
        return groups_indexed

    def unset_ike1_dh_group(self, ifname, group):
        self.delete_item("{}[name='{}']/ike-phase-1/dh-groups[group='{}']".format(self.namespace, ifname, group))

    def unset_dh_groups(self, phase, ifname):
        self.delete_item("{}[name='{}']/{}/dh-groups".format(self.namespace, ifname, phase))

    def unset_ike1_dh_groups(self, ifname):
        self.delete_item("{}[name='{}']/ike-phase-1/dh-groups".format(self.namespace, ifname))

    # Define get, set and unset methods for IKE1 proposals
    def set_ike1_proposals(self, ifname, id):
        self.set_item("{}[name='{}']/ike-phase-1/proposals[id='{}']".format(self.namespace, ifname, id), id)

    def get_ike1_proposal(self, ifname, id):
        return self.get_item("{}[name='{}']/ike-phase-1/proposals[id='{}']".format(self.namespace, ifname, id))

    def get_ike1_proposal_param(self, ifname, id, param):
        return self.get_item("{}[name='{}']/ike-phase-1/proposals[id='{}']/{}".format(self.namespace, ifname, id, param))

    def get_proposal_param(self, phase, ifname, id, param):
        return self.get_item("{}[name='{}']/{}/proposals[id='{}']/{}".format(self.namespace, ifname, phase, id, param))

    def set_proposal_param(self, phase, ifname, id, param, value):
        self.set_item("{}[name='{}']/{}/proposals[id='{}']/{}".format(self.namespace, ifname, phase, id, param), value)

    def get_ike1_proposals(self, name):
        proposals_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'ike-phase-1' in interface:
                    for config in interface['ike-phase-1']:
                        if 'proposals' in config:
                            proposals_indexed = interface['ike-phase-1']['proposals']
        return proposals_indexed

    def unset_proposal(self, phase, ifname, id):
        self.delete_item("{}[name='{}']/{}/proposals[id='{}']".format(self.namespace, ifname, phase, id))

    def set_proposal(self, phase, ifname, id):
        self.set_item("{}[name='{}']/{}/proposals[id='{}']".format(self.namespace, ifname, phase, id), id)

    def unset_ike1_proposal(self, ifname, id):
        self.delete_item("{}[name='{}']/ike-phase-1/proposals[id='{}']".format(self.namespace, ifname, id))

    def unset_ike1_proposals(self, ifname):
        self.delete_item("{}[name='{}']/ike-phase-1/proposals".format(self.namespace, ifname))

    def set_ike1_proposals_param(self, ifname, id, param, value):
        self.set_item("{}[name='{}']/ike-phase-1/proposals[id='{}']/{}".format(self.namespace, ifname, id, param), value)

    # Define get, set and unset methods for IKE2 local and remote traffic selectors
    def set_ike2_local_ts(self, ifname, value):
        self.set_item("{}[name='{}']/ike-phase-2/local-ts".format(self.namespace, ifname), value)

    def get_ike2_local_ts(self, ifname):
        return self.get_item("{}[name='{}']/ike-phase-2/local-ts".format(self.namespace, ifname))

    def unset_ike2_local_ts(self, ifname):
        self.delete_item("{}[name='{}']/ike-phase-2/local-ts".format(self.namespace, ifname))

    def set_ike2_remote_ts(self, ifname, value):
        self.set_item("{}[name='{}']/ike-phase-2/remote-ts".format(self.namespace, ifname), value)

    def get_ike2_remote_ts(self, ifname):
        return self.get_item("{}[name='{}']/ike-phase-2/remote-ts".format(self.namespace, ifname))

    def unset_ike2_remote_ts(self, ifname):
        self.delete_item("{}[name='{}']/ike-phase-2/remote-ts".format(self.namespace, ifname))

    # Define get, set and unset methods for IKE2 Diffie-Hellman groups
    def set_ike2_dh_groups(self, ifname, group):
        try:
            self.set_item("{}[name='{}']/ike-phase-2/dh-groups[group='{}']".format(self.namespace, ifname, group), group)
        except Exception as ex:
            self.logger.error('Set Diffie-Hellman groups "{group}" failed. Not valid group. {msg}'.format(group=group, msg=str(ex)))
            print('Set Diffie-Hellman groups "{group}" failed. Not valid group. {msg}'.format(group=group, msg=str(ex)))

    def get_ike2_dh_groups(self, name):
        groups_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'ike-phase-2' in interface:
                    for config in interface['ike-phase-2']:
                        if 'dh-groups' in config:
                            groups_indexed = interface['ike-phase-2']['dh-groups']
        return groups_indexed

    def unset_ike2_dh_group(self, ifname, group):
        self.delete_item("{}[name='{}']/ike-phase-2/dh-groups[group='{}']".format(self.namespace, ifname, group))

    def unset_ike2_dh_groups(self, ifname):
        self.delete_item("{}[name='{}']/ike-phase-2/dh-groups".format(self.namespace, ifname))

    # Define get, set and unset methods for IKE2 proposals
    def set_ike2_proposals(self, ifname, id):
        self.set_item("{}[name='{}']/ike-phase-2/proposals[id='{}']".format(self.namespace, ifname, id), id)

    def get_ike2_proposal(self, ifname, id):
        return self.get_item("{}[name='{}']/ike-phase-2/proposals[id='{}']".format(self.namespace, ifname, id))

    def get_ike2_proposal_param(self, ifname, id, param):
        return self.get_item("{}[name='{}']/ike-phase-2/proposals[id='{}']/{}".format(self.namespace, ifname, id, param))

    def get_ike2_proposals(self, name):
        proposals_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'ike-phase-2' in interface:
                    for config in interface['ike-phase-2']:
                        if 'proposals' in config:
                            proposals_indexed = interface['ike-phase-2']['proposals']
        return proposals_indexed

    def unset_ike2_proposal(self, ifname, id):
        self.delete_item("{}[name='{}']/ike-phase-2/proposals[id='{}']".format(self.namespace, ifname, id))

    def unset_ike2_proposals(self, ifname):
        self.delete_item("{}[name='{}']/ike-phase-2/proposals".format(self.namespace, ifname))

    def set_ike2_proposals_param(self, ifname, id, param, value):
        self.set_item("{}[name='{}']/ike-phase-2/proposals[id='{}']/{}".format(self.namespace, ifname, id, param), value)

    def get_config_mode(self, ifname):
        return self.get_item("{}[name='{}']/config-mode".format(self.namespace, ifname))

    def set_config_mode(self, ifname, value):
        self.set_item("{}[name='{}']/config-mode".format(self.namespace, ifname), value)

    def get_install_vip(self, ifname):
        return self.get_item("{}[name='{}']/install-virtual-ip".format(self.namespace, ifname))

    def set_install_vip(self, ifname, value):
        self.set_item("{}[name='{}']/install-virtual-ip".format(self.namespace, ifname), value)

    def unset_install_vip(self, ifname):
        self.delete_item("{}[name='{}']/install-virtual-ip".format(self.namespace, ifname))

    def get_vip(self, ifname):
        return self.get_item("{}[name='{}']/virtual-ip".format(self.namespace, ifname))

    def set_vip(self, ifname, value):
        self.set_item("{}[name='{}']/virtual-ip".format(self.namespace, ifname), value)

    def unset_vip(self, ifname):
        self.delete_item("{}[name='{}']/virtual-ip".format(self.namespace, ifname))

    def get_if_status(self, ifname):
        data = dict()
        xpath = "{}[name='{}']".format(self.status_namespace, ifname)

        values = self.session.get_items(xpath + "//*")
        if not values:
            return data
        for i in range(values.val_cnt()):
            val = self.get_data_value(values.val(i))
            data[values.val(i).xpath()] = val

        return data

    def set_interface_wifi_securty_param(self, name, param, value):
        self.set_item("{}[name='{}']/security/{}".format(self.namespace, name, param), value)

    def get_interface_wifi_securty_param(self, param):
        return self.get_item("/ucpe:config/interfaces:interface[name='wifi0']/security/{}".format(param))

    def unset_interface_wifi_securty_param(self, name, param):
        self.delete_item("{}[name='{}']/security/{}".format(self.namespace, name, param))

    def set_interface_wifi_radio_param(self, name, param, value):
        self.set_item("{}[name='{}']/radio-settings/{}".format(self.namespace, name, param), value)

    def get_interface_wifi_radio_param(self, name, param, default=None):
        try:
            for interface in self.get_running_config(full_path=False):
                if interface['name'] == name and param in interface:
                    return self.get_item("{}[name='{}']/radio-settings/{}".format(self.namespace, name, param))
        except:
            return default

    def unset_interface_wifi_radio_param(self, name, param):
        self.delete_item("{}[name='{}']/radio-settings/{}".format(self.namespace, name, param))

    def set_wifi_wan_interface(self, wifi_name, name):
        self.set_item("{}[name='{}']/wifi-wan-interface[name='{}']".format(self.namespace, wifi_name, name), name)

    def get_wifi_wan_interfaces(self, name):
        config_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'wifi-wan-interfacee' in interface:
                    for wan in interface['wifi-wan-interface']:
                        config_indexed[wan['name']] = wan
        return config_indexed

    def unset_wifi_wan_interface(self, wifi_name, name):
        self.delete_item("{}[name='{}']/wifi-wan-interface[name='{}']".format(self.namespace, wifi_name, name))

    def unset_wifi_wan_interfaces(self, wifi_name):
        self.delete_item("{}[name='{}']/wifi-wan-interface".format(self.namespace, wifi_name))

    def set_mac_allow(self, ifname, mac):
        self.set_item("{}[name='{}']/security/mac-allow[mac='{}']".format(self.namespace, ifname, mac), mac)

    def get_mac_allow(self, ifname, mac):
        return self.get_item("{}[name='{}']/security/mac-allow[mac='{}']".format(self.namespace, ifname, mac))

    def get_mac_allow_list(self, name):
        macs = []
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'mac-allow' in interface:
                    macs = interface['mac-allow']
                break
        return macs

    def unset_mac_allow(self, ifname, mac):
        self.delete_item("{}[name='{}']/security/mac-allow[mac='{}']".format(self.namespace, ifname, mac))

    def unset_mac_allow_list(self, ifname):
        self.delete_item("{}[name='{}']/security/mac-allow".format(self.namespace, ifname))

    def set_mac_deny(self, ifname, mac):
        self.set_item("{}[name='{}']/security/mac-deny[mac='{}']".format(self.namespace, ifname, mac), mac)

    def get_mac_deny(self, ifname, mac):
        return self.get_item("{}[name='{}']/security/mac-deny[mac='{}']".format(self.namespace, ifname, mac))

    def get_mac_deny_list(self, name):
        macs = []
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'mac-deny' in interface:
                    macs = interface['mac-deny']
                break
        return macs

    def unset_mac_deny(self, ifname, mac):
        self.delete_item("{}[name='{}']/security/mac-deny[mac='{}']".format(self.namespace, ifname, mac))

    def unset_mac_deny_list(self, ifname):
        self.delete_item("{}[name='{}']/security/mac-deny".format(self.namespace, ifname))

    def get_container_param(self, ifname, container, param, default=None):
        try:
            interface = next((intf for intf in self.get_running_config(full_path=False) if intf['name'] == ifname), None)
            if interface is None:
                return default
            else:
                return self.get_item("{}[name='{}']/{}/{}".format(self.namespace, ifname, container, param))
        except:
            return default

    def set_container_param(self, ifname, container, param, value):
        self.set_item("{}[name='{}']/{}/{}".format(self.namespace, ifname, container, param), value)

    def unset_container_param(self, ifname, container, param):
        self.delete_item("{}[name='{}']/{}/{}".format(self.namespace, ifname, container, param))

    def unset_tracehost_probes(self, ifname):
        self.delete_item("{}[name='{}']/tracehost/probes".format(self.namespace, ifname))

    def unset_tracehost_probe(self, ifname, id):
        self.delete_item("{}[name='{}']/tracehost/probes[id='{}']".format(self.namespace, ifname, id))

    def set_tracehost_probe(self, ifname, id):
        self.set_item("{}[name='{}']/tracehost/probes[id='{}']".format(self.namespace, ifname, id), id)

    def set_tracehost_probe_param(self, ifname, id, param, value):
        self.set_item("{}[name='{}']/tracehost/probes[id='{}']/{}".format(self.namespace, ifname, id, param), value)

    def unset_tracehost_probe_param(self, ifname, id, param):
        self.delete_item("{}[name='{}']/tracehost/probes[id='{}']/{}".format(self.namespace, ifname, id, param))

    def unset_wireguard_peers(self, ifname):
        self.delete_item("{}[name='{}']/peers".format(self.namespace, ifname))

    def unset_wireguard_peer(self, ifname, endpoint):
        self.delete_item("{}[name='{}']/peers[peer-endpoint='{}']".format(self.namespace, ifname, endpoint))

    def set_wireguard_peer(self, ifname, endpoint):
        self.set_item("{}[name='{}']/peers[peer-endpoint='{}']".format(self.namespace, ifname, endpoint), endpoint)

    def set_wireguard_peer_param(self, ifname, endpoint, param, value):
        self.set_item("{}[name='{}']/peers[peer-endpoint='{}']/{}".format(self.namespace, ifname, endpoint, param), value)

    def unset_wireguard_peer_param(self, ifname, endpoint, param):
        self.delete_item("{}[name='{}']/peers[peer-endpoint='{}']/{}".format(self.namespace, ifname, endpoint, param))

    def get_wireguard_peer_params(self, name, param, default=None):
        try:
            for interface in self.get_running_config(full_path=False):
                if interface['name'] == name and 'peers' in interface:
                    return self.get_item("{}[name='{}']/peers/{}".format('/ucpe:config/interfaces:interface', name, param))
        except:
            return default

    def get_wireguard_peer_all_params(self, name, default=None):
        try:
            for interface in self.get_running_config(full_path=False):
                if interface['name'] == name and 'peers' in interface:
                    return self.session.get_data("{}[name='{}']/peers".format('/ucpe:config/interfaces:interface', name))
        except:
            return default

    def set_wireguard_concentrator_param(self, ifname, param, value):
        self.set_item("{}[name='{}']/wireguard-concentrator/{}".format(self.namespace, ifname, param), value)

    def unset_wireguard_concentrator_param(self, ifname, param):
        self.delete_item("{}[name='{}']/wireguard-concentrator/{}".format(self.namespace, ifname, param))

    def del_wireguard_concentrator(self, name):
        self.delete_item("{}[name='{}']/wireguard-concentrator".format(self.namespace, name))

    def get_wireguard_concentrator_param(self, name, param, default=None):
        try:
            for interface in self.get_running_config(full_path=False):
                if interface['name'] == name and 'wireguard-concentrator' in interface:
                    return self.get_item("{}[name='{}']/wireguard-concentrator/{}".format(self.namespace, name, param))
        except:
            return default

    def get_wireguard_concentrator_params(self, name, default=None):
        try:
            for interface in self.get_running_config(full_path=False):
                if interface['name'] == name and 'wireguard-concentrator' in interface:
                    return self.session.get_data("{}[name='{}']/wireguard-concentrator".format(self.namespace, name))
        except:
            return default

    def get_tracehost_probes(self, name):
        config_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'tracehost' in interface:
                    if 'probes' in interface['tracehost']:
                        for probe in interface['tracehost']['probes']:
                            config_indexed[probe['id']] = probe
        return config_indexed

    def get_vxlan_rvtep(self, ifname):
        rvteps = []
        vals = self.session.get_items("{}[name='{}']/remote/vtep-addrs//*".format(self.namespace, ifname))
        if vals:
            for i in range(vals.val_cnt()):
                if vals.val(i).dflt():
                    continue
                rvteps.append(self.get_data_value(vals.val(i)))
        return rvteps

    def add_vxlan_rvtep(self, ifname, address):
        # get current list of vteps
        cvteps = self.get_vxlan_rvtep(ifname)
        # vals = self.session.get_items("{}[name='{}']/remote/vtep-addrs//*".format(self.namespace, ifname))
        # if vals:
        #     for i in range(vals.val_cnt()):
        #         if vals.val(i).dflt():
        #             continue

        #         cvteps.append(self.get_data_value(vals.val(i)))

        if address in cvteps:
            # allready in config
            return

        return self.set_item("{}[name='{}']/remote/vtep-addrs[vtep-addr='{}']".format(self.namespace, ifname, address), address)

    def del_vxlan_rvtep(self, ifname, address):
        # get current list of vteps
        cvteps = self.get_vxlan_rvtep(ifname)
        # vals = self.session.get_items("{}[name='{}']/remote/vtep-addrs//*".format(self.namespace, ifname))
        # if vals:
        #     for i in range(vals.val_cnt()):
        #         if vals.val(i).dflt():
        #             continue

        #         cvteps.append(self.get_data_value(vals.val(i)))

        if address not in cvteps:
            # allready in config
            return

        return self.delete_item("{}[name='{}']/remote/vtep-addrs[vtep-addr='{}']".format(self.namespace, ifname, address))

    def del_vxlan_rvteps(self, ifname):
        return self.delete_item("{}[name='{}']/remote/vtep-addrs".format(self.namespace, ifname))

    def handle_vpp_interface_addresses(self, ifname, ipv4_prefix=None, ipv6_prefix=None):
        """ ifname param - br-dataX
        """

        if self.get_network_mode() != 'vpp':
            return

        bvi_idx = None
        host_if = None
        host_idx = None
        
        self.logger.info("handle_vpp_interface_addresses - {} - {} - {}" .format(ifname, ipv4_prefix, ipv6_prefix))
        vpp_if = vppif.VppApiInterface('if-addresses')
        host_if = vpp_if.get_tap_interface_by_host_ifname(ifname)
        if not host_if:
            del(vpp_if)
            return
        bridge = vpp_if.get_interface_bd_id(host_if['sw_if_index'])
        if bridge:
            bvi_idx = vpp_if.get_bd_bvi_by_bdidx(bridge)
        host_idx = host_if['sw_if_index']

        if not bvi_idx or not host_idx:
            del(vpp_if)
            return

        if ipv4_prefix:
            ipv4_a = ipaddress.ip_interface(ipv4_prefix)
            ret = vpp_if.add_interface_address(idx=bvi_idx,addr=ipv4_prefix)
            if ret:
                self.logger.error(" Error adding interface {} address {} in vpp - {} ". format(ifname,ipv4_prefix,ret))
                del(vpp_if)
                return
            ret = vpp_if.drctl_set_ip( host_idx, ipv4_a.ip.packed )
            self.logger.info("handle_vpp_interface_addresses {} ipv4 - {} added" .format(ifname, ipv4_a.packed))

        if ipv6_prefix:
            ipv6_a = ipaddress.ip_interface(ipv6_prefix)
            ret = vpp_if.add_interface_address(idx=bvi_idx,addr=ipv6_prefix)
            if ret:
                self.logger.error(" Error adding interface {} address {} in vpp - {} ". format(ifname,ipv6_prefix,ret))
                del(vpp_if)
                return
            ret = vpp_if.drctl_set_ip( host_idx, ipv6_a.ip.packed )
            vpp_if.ip6_ra_config(bvi_idx, suppress=1)
            self.logger.info("handle_vpp_interface_addresses {} ipv6 - {} added" .format(ifname, ipv6_a.ip.packed))

        del(vpp_if)
        return

    def handle_vpp_vlan_interface(self, name, change, operation):
        if self.get_network_mode() != 'vpp':
            return
        vpp_if = vppif.VppApiInterface('sr-interfaces')
        vpp_iface = name
        ifs_to_transfer = []
        target_bd = None
        bvi_idx = None
        L2_API_PORT_TYPE_BVI = 1
        vpp_iface = "host-" + vpp_iface if self.get_interface_param(name, 'mode') == 'virtio' else vpp_iface
        if operation == 'CREATED':
            if change['new_value'] is not None and int(change['new_value']) != 0:
                (ifs_to_transfer, bvi_idx) = self.get_ifs_to_transfer(vpp_if, vpp_iface, name, 0)
                target_bd = self.create_vlan_bd(vpp_if, name, vpp_if.get_ifidx_by_name(vpp_iface), int(change['new_value']))
        if operation == 'MODIFIED':
            vpp_iface = self.tag_interface(vpp_iface, int(change['old_value']))
            (ifs_to_transfer, bvi_idx) = self.get_ifs_to_transfer(vpp_if, vpp_iface, name, int(change['old_value']))
            vpp_iface = vpp_iface.replace('.' + change['old_value'], '')
            target_bd = self.create_vlan_bd(vpp_if, name, vpp_if.get_ifidx_by_name(vpp_iface), int(change['new_value']))
            vpp_iface = self.tag_interface(vpp_iface, int(change['old_value']))
            ret = vpp_if.delete_vlan_subinterface(vpp_if.get_ifidx_by_name(vpp_iface))
            vpp_iface = vpp_iface.replace('.' + change['old_value'], '')
        if operation == 'DELETED':
            vpp_iface = self.tag_interface(vpp_iface, int(change['old_value']))
            (ifs_to_transfer, bvi_idx) = self.get_ifs_to_transfer(vpp_if, vpp_iface, name, int(change['old_value']))
            ret = vpp_if.delete_vlan_subinterface(vpp_if.get_ifidx_by_name(vpp_iface))
            target_bd = vpp_if.get_ifidx_by_name(vpp_iface.replace('.' + change['old_value'], ''))
            vpp_iface = vpp_iface.replace('.' + change['old_value'], '')

        vlan = 0 if operation == 'DELETED' else int(change['new_value'])

        # Transfer interfaces to target bridge-domain
        if target_bd is not None and bvi_idx is not None:
            for iface_idx in ifs_to_transfer:
                if bvi_idx == iface_idx:
                    vpp_if.add_to_bridge(target_bd, iface_idx, L2_API_PORT_TYPE_BVI)
                    continue
                vpp_if.add_to_bridge(target_bd, iface_idx)
                vpp_if.set_interface_learning(iface_idx, disable=1)

        # Delete bridge-domains which are empty or ones with single irrelevant sub-interface
        bridge_domains = vpp_if.get_bridge_domains()
        for bd in bridge_domains:
            if bd['n_sw_ifs'] == 0:
                ret = vpp_if.delete_bridge_domain(bd['bd_id'])
                self.logger.debug("Deleting bridge domain with ID {} returned {}".format(bd['bd_id'], ret))
            if bd['n_sw_ifs'] == 1:
                if_details = bd['sw_if_details']
                ifname = vpp_if.get_ifname_by_ifidx(if_details[0]['sw_if_index'])
                if vpp_if.is_vlan_sub_interface(ifname):
                    ret = vpp_if.delete_vlan_subinterface(vpp_if.get_ifidx_by_name(ifname))
                    ret = vpp_if.delete_bridge_domain(bd['bd_id'])
                    self.logger.debug("Deleting bridge domain with ID {} returned {}".format(bd['bd_id'], ret))

        del(vpp_if)

        # Handle PPPoE interfaces associated with this interface
        vpp_iface = self.tag_interface(vpp_iface, vlan)
        # get only dataX from a.ethif as it could have been tagged
        ppp_if = next((a for a in self.pppoe_interfaces if a.ethif is not None and a.ethif.split('.')[0].replace('host-', '') == name),
                      None)
        if ppp_if:
            ppp_if.ethif = vpp_iface
        else:
            self.logger.debug("No PPPoE interfaces associated with this interface")

    def create_vlan_bd(self, vpp_if, ifname, ifid, vlan):
        # Check if sub-interface with this vlan is already created by neutron.
        is_created = vpp_if.get_ifidx_by_name(ifname + '.' + str(vlan))
        if is_created is not None:
            return is_created
        (retval, subidx) = vpp_if.create_vlan_subinterface(idx=ifid, vlan_id=vlan)
        if retval == 0 and vpp_if.vlan_rewrite_poptag1(subidx) == 0:
            ret = vpp_if.create_bridge_domain(subidx, tag='br-{}.{}'.format(ifname, vlan))
            vpp_if.add_to_bridge(subidx, subidx)
            vpp_if.ifup(subidx)
            return subidx

    def tag_interface(self, iface, vlan):
        if vlan is not None and vlan != 0:
            iface += '.' + str(vlan)
            return iface
        return iface

    def get_ifs_to_transfer(self, vpp_if, vpp_iface, name, vlan):
        ifs_to_transfer = []
        bd_tag = 'br-{}'.format(name)
        if vlan:
            bd_tag += '.' + str(vlan)
        identifier = name.replace('data', '')
        host_iface_idx = vpp_if.get_ifidx_by_name(vpp_iface)
        bd = vpp_if.get_bd_by_tag(tag=bd_tag)
        if not bd:
            return ifs_to_transfer
        bvi_idx = bd.get('bvi_sw_if_index', None)
        bd_ifs = list(map(lambda x: x['sw_if_index'], bd['sw_if_details']))
        for idx in bd_ifs:
            if idx == host_iface_idx:
                continue
            if vpp_if.get_ifname_by_ifidx(idx) != 'tap' + identifier and 'data' not in vpp_if.get_ifname_by_ifidx(idx) and idx != bvi_idx:
                continue
            ifs_to_transfer.append(idx)

        return (ifs_to_transfer, bvi_idx)

    def _get_wireguard_concentrator_param(self, config, param):
        return config.get('wireguard-concentrator', {}).get(param, '')

    def get_wireguard_peer_all_params(self, name, default=None):
        try:
            for interface in self.get_running_config(full_path=False):
                if interface['name'] == name and 'peers' in interface:
                    return self.session.get_data("{}[name='{}']/peers".format('/ucpe:config/interfaces:interface', name))
        except:
            return default

    def _get_local_wireguard_endpoint_ip(self, endpoint):
        if 'data' in endpoint or 'lte' in endpoint or 'wifi' in endpoint:
                    bootproto = None
                    address = None
                    addressv6 = None
                    bootproto = self.get_item("/ucpe:config/interfaces:interface[name='{}']/bootproto".format(endpoint))
                    if bootproto == "none" and bootproto != None:
                        addressv6 = self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv6-prefix".format(endpoint))
                        if addressv6 is not None:
                            n_localv6 = ipaddress.ip_interface(addressv6).ip

                        address = self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv4-prefix".format(endpoint))
                        if address is not None:
                            n_local = ipaddress.ip_interface(address).ip
        elif endpoint != None and endpoint != "":
            n_local = ipaddress.ip_address(endpoint) if isinstance(ipaddress.ip_address(endpoint), ipaddress.IPv4Address) else None
            n_localv6 = ipaddress.ip_address(endpoint) if isinstance(ipaddress.ip_address(endpoint), ipaddress.IPv6Address) else None

        return (n_local, n_localv6)

